import os


def test_main(capfd):
    exit_status = os.system("locan --help")
    captured = capfd.readouterr()
    assert captured.out.startswith("usage:")
    assert exit_status == 0

    exit_status = os.system("locan --version")
    captured = capfd.readouterr()
    assert captured.out.startswith("locan version")
    assert exit_status == 0

    exit_status = os.system("locan --info")
    captured = capfd.readouterr()
    for item in ["Locan:", "Python:", "System:", "Python dependencies:"]:
        assert item in captured.out
    assert exit_status == 0



#!/usr/bin/env python

"""
Show system information and dependency versions.


To run the script::

    locan show_versions -v -e -o <module name> [<module name>...]

Try for instance::

    locan show_versions -v -e

See Also
--------
locan.utils.system_information.show_versions
"""
import argparse

from locan.utils.system_information import show_versions as sc_show_versions


def _add_arguments(parser):
    parser.add_argument(
        "-v",
        "--verbose",
        dest="verbose",
        action="store_true",
        help="Include information on node and executable path.",
    )
    parser.add_argument(
        "-e",
        "--extra",
        dest="extra",
        action="store_true",
        help="Include extra dependencies as specified in setup.py.",
    )
    parser.add_argument(
        "-o",
        "--other",
        dest="other",
        type=str,
        nargs="*",
        help="Include other module names.",
    )


def main(args=None):
    parser = argparse.ArgumentParser(
        description="Show system information and dependency versions."
    )
    _add_arguments(parser)
    returned_args = parser.parse_args(args)
    sc_show_versions(
        verbose=returned_args.verbose,
        extra_dependencies=returned_args.extra,
        other_dependencies=returned_args.other,
    )


if __name__ == "__main__":
    main()



#!/usr/bin/env python

"""
Define regions of interest with napari and save as roi files.

With this script you can choose a file name, open the localization file in napari.
Draw regions of interest as additional shapes in napari.
Upon closing napari each shape is taken as single roi and saved as _roi.yaml file.

To run the script::

    locan rois -f <file> -t <file type> -i <roi file indicator> --bin_size <>
    --rescale <string with tuple or rescale>

Try for instance::

    locan rois -f "locan/tests/test_data/five_blobs.txt" -t 1 -i "_roi" --bin_size 10

See Also
--------
locan.data.rois.select_by_drawing_napari
"""
import argparse
import re
from pathlib import Path

import locan.locan_io.locdata.io_locdata as io
from locan.constants import FileType
from locan.dependencies import HAS_DEPENDENCY
from locan.gui.io import file_dialog
from locan.visualize.napari.utilities import select_by_drawing_napari

if HAS_DEPENDENCY["napari"]:
    import napari


def sc_draw_roi_napari(
    file_path=None, file_type=FileType.CUSTOM, roi_file_indicator="_roi", **kwargs
):
    """
    Define regions of interest by drawing a boundary.

    Parameters
    ----------
    file_path : str, bytes, os.PathLike
        File path to localization data.
    file_type : int, str, locan.constants.FileType, metadata_pb2
        Indicator for the file type.
        Integer or string should be according to locan.constants.FileType.
    roi_file_indicator : str
        Indicator to add to the localization file name and use as roi file name (with further extension .yaml).
    kwargs : dict
        Other parameters passed to :func:`render_2d_napari`.
    """
    # choose file interactively
    if file_path is None:
        file_path = Path(file_dialog(message="choose file", filter="*.txt; *.csv")[0])
    else:
        file_path = Path(file_path)

    print(file_path)

    # load data
    dat = io.load_locdata(path=file_path, file_type=file_type)

    # set roi
    rois = select_by_drawing_napari(locdata=dat, **kwargs)
    napari.run()
    print(rois)

    # save roi
    roi_path_list = []
    for i, roi in enumerate(rois):
        roi_file = file_path.stem + roi_file_indicator + f"_{i}.yaml"
        roi_path = file_path.with_name(roi_file)
        roi_path_list.append(roi_path)
        roi.to_yaml(path=roi_path)

    return roi_path_list


def type_converter_rescale(input_string):
    if input_string == "None":
        return None
    elif input_string == "True":
        return True
    elif input_string == "False":
        return False
    else:
        pattern = re.match(
            r"\(?([0-9]*[.]?[0-9]+),?\s?([0-9]*[.]?[0-9]+)\)?", input_string
        )
        if pattern:
            return tuple(float(element) for element in pattern.groups())
        else:
            return input_string


def _add_arguments(parser):
    parser.add_argument(
        "-f",
        "--file",
        dest="file",
        type=str,
        default=None,
        help="File path to localization data.",
    )
    parser.add_argument(
        "-t",
        "--type",
        dest="type",
        type=int,
        default=2,
        help="Integer or string indicating the file type.",
    )
    parser.add_argument(
        "-i",
        "--indicator",
        dest="roi_file_indicator",
        type=str,
        default="_roi",
        help="Indicator to add to the localization file name and use as roi file name "
        "(with further extension .yaml).",
    )
    parser.add_argument(
        "--bin_size",
        dest="bin_size",
        type=float,
        default=50,
        help="Keyword passed to render function.",
    )
    parser.add_argument(
        "--rescale",
        dest="rescale",
        type=type_converter_rescale,
        default="EQUALIZE",
        help="Rescale intensity values. Keyword passed to render_2d_napari function.",
    )


def main(args=None):

    parser = argparse.ArgumentParser(description="Set roi by drawing a boundary.")
    _add_arguments(parser)
    returned_args = parser.parse_args(args)

    sc_draw_roi_napari(
        returned_args.file,
        returned_args.type,
        returned_args.roi_file_indicator,
        bin_size=returned_args.bin_size,
        rescale=returned_args.rescale,
    )


if __name__ == "__main__":
    main()



"""

File input/output for localization data.

"""
from __future__ import annotations

import logging
from enum import Enum

import pandas as pd

import locan.constants
from locan.data import metadata_pb2
from locan.data.locdata import LocData
from locan.locan_io.locdata.asdf_io import load_asdf_file
from locan.locan_io.locdata.decode_io import load_decode_file
from locan.locan_io.locdata.elyra_io import load_Elyra_file
from locan.locan_io.locdata.nanoimager_io import load_Nanoimager_file
from locan.locan_io.locdata.rapidstorm_io import (
    load_rapidSTORM_file,
    load_rapidSTORM_track_file,
)
from locan.locan_io.locdata.smap_io import load_SMAP_file
from locan.locan_io.locdata.smlm_io import load_SMLM_file
from locan.locan_io.locdata.thunderstorm_io import load_thunderstorm_file
from locan.locan_io.locdata.utilities import (
    convert_property_names,
    convert_property_types,
)

__all__ = ["load_txt_file", "load_locdata"]

logger = logging.getLogger(__name__)


def load_txt_file(
    path,
    sep=",",
    columns=None,
    nrows=None,
    property_mapping=None,
    convert=True,
    **kwargs,
):
    """
    Load localization data from a txt file.

    Locan column names are either supplied or read from the first line header.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a localization file to load.
    sep : str
        separator between column values (Default: ',')
    columns : list of str, None
        Locan column names. If None the first line is interpreted as header (Default: None).
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded (Default: None).
    property_mapping : dict[str: str] or list[dict]
        Mappings between column names and locan property names
    convert : bool
        If True convert types by applying type specifications in locan.constants.PROPERTY_KEYS.
    kwargs : dict
        Other parameters passed to `pandas.read_csv()`.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    # define columns
    if columns is None:
        dataframe = pd.read_csv(
            path, sep=sep, nrows=nrows, **dict(dict(skiprows=0), **kwargs)
        )

        column_keys = convert_property_names(
            dataframe.columns, property_mapping=property_mapping
        )
        dataframe.columns = column_keys
    else:
        column_keys = convert_property_names(columns, property_mapping=property_mapping)
        dataframe = pd.read_csv(
            path,
            sep=sep,
            nrows=nrows,
            **dict(dict(skiprows=1, names=column_keys), **kwargs),
        )

    if convert:
        dataframe = convert_property_types(
            dataframe, types=locan.constants.PROPERTY_KEYS
        )

    dat = LocData.from_dataframe(dataframe=dataframe)

    dat.meta.source = metadata_pb2.EXPERIMENT
    dat.meta.state = metadata_pb2.RAW
    dat.meta.file.type = metadata_pb2.CUSTOM
    dat.meta.file.path = str(path)

    del dat.meta.history[:]
    dat.meta.history.add(
        name="load_txt_file",
        parameter="path={}, sep={}, columns={}, nrows={}".format(
            path, sep, columns, nrows
        ),
    )

    return dat


def _map_file_type_to_load_function(file_type):
    """
    Interpret user input for file_type.

    Parameters
    ----------
    file_type : int, str, locan.constants.FileType, locan.data.metadata_pb2.Metadata
        Identifier for the file type. Integer or string should be according to locan.constants.FileType.

    Returns
    -------
    callable
        Name of function for loading the localization file of `type`.
    """
    look_up_table = dict(
        load_txt_file=load_txt_file,
        load_rapidSTORM_file=load_rapidSTORM_file,
        load_Elyra_file=load_Elyra_file,
        load_thunderstorm_file=load_thunderstorm_file,
        load_asdf_file=load_asdf_file,
        load_Nanoimager_file=load_Nanoimager_file,
        load_rapidSTORM_track_file=load_rapidSTORM_track_file,
        load_SMLM_file=load_SMLM_file,
        load_decode_file=load_decode_file,
        load_SMAP_file=load_SMAP_file,
    )

    class LoadFunction(Enum):
        load_txt_file = 1
        load_rapidSTORM_file = 2
        load_Elyra_file = 3
        load_thunderstorm_file = 4
        load_asdf_file = 5
        load_Nanoimager_file = 6
        load_rapidSTORM_track_file = 7
        load_SMLM_file = 8
        load_decode_file = 9
        load_SMAP_file = 10

    try:
        if isinstance(file_type, int):
            function_name = LoadFunction(file_type).name
        elif isinstance(file_type, str):
            function_name = LoadFunction(
                locan.constants.FileType[file_type.upper()].value
            ).name
        elif isinstance(file_type, locan.constants.FileType):
            function_name = LoadFunction(file_type.value).name
        elif isinstance(file_type, metadata_pb2):
            function_name = LoadFunction(file_type).name
        else:
            raise TypeError
        return look_up_table[function_name]
    except ValueError:
        raise ValueError(f"There is no load function for type {file_type}.")


def load_locdata(path, file_type=1, nrows=None, **kwargs):
    """
    Load data from localization file as specified by type.

    This function is a wrapper for read functions for the various types of SMLM data.

    Parameters
    ----------
    path : str, bytes, os.PathLike, file-like
        File path for a localization data file to load.
    file_type : int, str, locan.constants.FileType, locan.data.metadata_pb2.Metadata
        Indicator for the file type.
        Integer or string should be according to locan.constants.FileType.
    nrows : int, None
        The number of localizations to load from file. None means that all available rows are loaded.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations.
    """
    return _map_file_type_to_load_function(file_type)(path=path, nrows=nrows, **kwargs)



"""
Module to deal with required and optional dependencies.

Optional dependencies are defined in setup.cfg as [options.extras_require].

In any module that requires an optional dependency the import should be carried out on top:
    if HAS_DEPENDENCY["package"]: import package

Any function that makes use of the optional dependency should be decorated with
    @needs_package("package")


CONSTANTS

.. autosummary::
   :toctree: ./

   INSTALL_REQUIRES
   EXTRAS_REQUIRE
   IMPORT_NAMES
   HAS_DEPENDENCY
"""
from __future__ import annotations

import importlib.metadata
import importlib.util
import logging
import os
import re
import sys
from collections.abc import Iterable
from enum import Enum
from functools import wraps

logger = logging.getLogger(__name__)

__all__ = [
    "needs_package",
    "IMPORT_NAMES",
    "INSTALL_REQUIRES",
    "EXTRAS_REQUIRE",
    "HAS_DEPENDENCY",
    "QtBindings",
]


def _has_dependency_factory(
    packages: Iterable[str], import_names: dict[str, str] | None = None
) -> dict[str, bool]:
    if import_names is None:
        import_names = IMPORT_NAMES
    has_dependency = dict()
    for package in packages:
        key = import_names.get(package, package)
        value = importlib.util.find_spec(key) is not None
        has_dependency[key] = value
    return has_dependency


def needs_package(package, import_names=None, has_dependency=None):
    """
    Function that returns a decorator to check for optional dependency.

    Parameters
    ----------
    package : str
        Package or dependency name that needs to be imported.
    import_names : dict[str, str] | None
        Mapping of package names onto import names.
    has_dependency : dict | None
        Dictionary with bool indicator if package (import name) is available.

    Returns
    -------
    callable
        A decorator that raises ImportError if package is not available.
    """
    if import_names is None:
        import_names = IMPORT_NAMES
    if has_dependency is None:
        has_dependency = HAS_DEPENDENCY
    import_name = import_names.get(package, package)

    def decorator(func):
        @wraps(func)
        def wrapper(*args, **kwargs):
            if not has_dependency[import_name]:
                raise ImportError(
                    f"Function {func} needs {import_name} which cannot be imported."
                )
            return func(*args, **kwargs)

        return wrapper

    return decorator


def _get_dependencies(package: str) -> tuple[set[str], set[str]]:
    """
    Read out all required and optional (extra) dependencies for locan.

    Returns
    -------
    tuple[set[str], set[str]]
        PyPI names
    """
    requires = importlib.metadata.requires(package)

    pattern = r"[\w-]+"
    required_dependencies = {
        re.match(pattern, item).group() for item in requires if "extra ==" not in item
    }
    extra_dependencies = {
        re.match(pattern, item).group() for item in requires if "extra ==" in item
    }

    return required_dependencies, extra_dependencies


#: List of required dependencies (PyPi package names)
INSTALL_REQUIRES = set()

#: List of optional dependencies (PyPi package names)
EXTRAS_REQUIRE = set()
INSTALL_REQUIRES, EXTRAS_REQUIRE = _get_dependencies(package="locan")

#: A dictionary mapping PyPi package names to import names if they are different
IMPORT_NAMES = dict()
IMPORT_NAMES["ruamel"] = "ruamel.yaml"
IMPORT_NAMES["fast-histogram"] = "fast_histogram"
IMPORT_NAMES["boost-histogram"] = "boost_histogram"
IMPORT_NAMES["protobuf"] = "google.protobuf"
IMPORT_NAMES["scikit-image"] = "skimage"
IMPORT_NAMES["scikit-learn"] = "sklearn"
IMPORT_NAMES["pytest-qt"] = "pytestqt"
IMPORT_NAMES["mpl-scatter-density"] = "mpl_scatter_density"

#: A dictionary indicating if dependency is available.
HAS_DEPENDENCY = _has_dependency_factory(
    packages=INSTALL_REQUIRES.union(EXTRAS_REQUIRE)
)


# Possible python bindings to interact with QT
class QtBindings(Enum):
    """
    Python bindings to interact with Qt.
    """

    NONE = ""
    PYSIDE2 = "pyside2"
    PYQT5 = "pyqt5"
    PYSIDE6 = "pyside6"
    PYQT6 = "pyqt6"


def _set_qt_binding(qt_binding: QtBindings | str) -> str:
    """
    Check if qtpy can import `qt_binding` and return the qt_binding that will be used.
    Checks os.environ["QT_API"] first.
    If os.environ["QT_API"] is set it will take precedence over `qt_bindings`.

    Note
    -----
    This function must be used before qtpy is imported for the first time
    to be effective.

    Returns
    -------
    str
        qt-bindings or empty string
    """
    if isinstance(qt_binding, QtBindings):
        qt_api = qt_binding.value
    elif qt_binding:
        qt_api = qt_binding.lower()
    else:
        qt_api = ""

    if "qtpy" in sys.modules:
        logger.warning("qtpy has already been loaded. QT_BINDING cannot be changed.")
    elif os.environ.get("QT_API", ""):
        qt_api = os.environ["QT_API"]
    else:
        if qt_api:
            os.environ["QT_API"] = qt_api

    try:
        from qtpy import (  # noqa: F401  # import API alone does not raise ModuleNotFoundError if no module is available
            API,
            QtCore,
        )

        if qt_api and qt_api != API:
            logger.warning(
                f"QT_BINDING {qt_api} is not available - {API} is used instead."
            )
            os.environ["QT_API"] = API
        qt_api = API
    except ModuleNotFoundError:
        if qt_api:
            logger.warning("QT_BINDING is not available.")
        qt_api = ""

    return qt_api



from copy import deepcopy

import matplotlib.pyplot as plt
import numpy as np
import pytest

import locan.constants
from locan import bunwarp, render_2d_mpl, render_2d_rgb_mpl
from locan.data.cluster import cluster_dbscan
from locan.data.region import Polygon
from locan.data.transform import overlay, randomize, transform_affine
from locan.data.transform.bunwarpj import _read_matrix, _unwarp
from locan.data.transform.spatial_transformation import _homogeneous_matrix
from locan.dependencies import HAS_DEPENDENCY
from locan.locan_io.locdata.io_locdata import load_asdf_file


def test_randomize_2d(locdata_2d):
    locdata_2d = deepcopy(locdata_2d)

    locdata_randomized = randomize(locdata_2d, hull_region="bb")
    # locdata_randomized.print_meta()
    assert len(locdata_randomized) == 6
    # print(locdata_randomized.meta)
    assert locdata_randomized.meta.history[-1].name == "randomize"

    locdata_randomized = randomize(locdata_2d, hull_region="ch")
    assert len(locdata_randomized) == 6

    locdata_randomized = randomize(locdata_2d, hull_region="obb")
    assert len(locdata_randomized) == 6

    locdata_2d.update_alpha_shape(alpha=10)
    locdata_randomized = randomize(locdata_2d, hull_region="as")
    assert len(locdata_randomized) == 6

    region = Polygon(((0, 0), (0, 5), (4, 3), (2, 0.5), (0, 0)))
    locdata_randomized = randomize(locdata_2d, hull_region=region)
    assert len(locdata_randomized) == 6


def test_randomize_3d(locdata_3d):
    locdata_randomized = randomize(locdata_3d, hull_region="bb")
    assert len(locdata_randomized) == 6
    assert locdata_randomized.meta.history[-1].name == "randomize"

    # todo: implement make_csr in 3d
    with pytest.raises(NotImplementedError):
        locdata_randomized = randomize(locdata_3d, hull_region="ch")
        assert len(locdata_randomized) == 6

    # region_dict = dict(region='polygon', region_specs=((0, 0, 0), (0, 5, 0), (4, 3, 2), (2, 0.5, 2), (0, 0, 0)))
    # locdata_randomized = randomize(locdata_3d, hull_region=region_dict)
    # assert len(locdata_randomized) == 6


@pytest.mark.parametrize(
    "fixture_name, expected",
    [
        # ('locdata_empty', 0),
        # ('locdata_single_localization', 1),
        ("locdata_2d", 6),
        ("locdata_3d", 6),
        ("locdata_non_standard_index", 6),
    ],
)
def test_randomize_locdata_objects(
    locdata_empty,
    locdata_single_localization,
    locdata_2d,
    locdata_3d,
    locdata_non_standard_index,
    fixture_name,
    expected,
):
    locdata = eval(fixture_name)
    locdata_randomized = randomize(locdata, hull_region="bb")
    assert len(locdata_randomized) == expected


def test_bunwarp_raw_transformation():
    matrix_path = (
        locan.ROOT_DIR
        / "tests/test_data/transform/BunwarpJ_transformation_raw_green.txt"
    )
    dat_green = load_asdf_file(
        path=locan.ROOT_DIR / "tests/test_data/transform/rapidSTORM_beads_green.asdf"
    )

    matrix_x, matrix_y = _read_matrix(path=matrix_path)
    assert np.array_equal(matrix_x.shape, [140, 140])

    dat_green_flipped = transform_affine(
        dat_green, matrix=[[-1, 0], [0, 1]], offset=[1400, 0]
    )
    new_points = _unwarp(
        dat_green_flipped.coordinates, matrix_x, matrix_y, pixel_size=(10, 10)
    )
    assert len(new_points) == len(dat_green)

    dat_green_transformed = bunwarp(
        locdata=dat_green, matrix_path=matrix_path, pixel_size=(10, 10), flip=True
    )
    assert len(dat_green_transformed) == len(dat_green)
    assert np.array_equal(dat_green_transformed.coordinates, new_points)
    assert dat_green_transformed.meta.history[-1].name == "bunwarp"

    # for visual inspection
    dat_red = load_asdf_file(
        path=locan.ROOT_DIR / "tests/test_data/transform/rapidSTORM_beads_red.asdf"
    )

    fig, ax = plt.subplots(1, 1, figsize=(16, 8))
    render_2d_mpl(
        dat_red,
        ax=ax,
        bin_size=5,
        bin_range=((0, 1000), (0, 1400)),
        rescale=True,
        cmap="Reds",
    )
    render_2d_mpl(
        dat_green,
        ax=ax,
        bin_size=5,
        bin_range=((0, 1000), (0, 1400)),
        rescale=True,
        cmap="Greens",
    )
    render_2d_mpl(
        dat_green_transformed,
        ax=ax,
        bin_size=5,
        bin_range=((0, 1000), (0, 1400)),
        rescale=True,
        cmap="Blues",
        alpha=0.5,
    )
    #  plt.show()

    render_2d_rgb_mpl(
        [dat_red, dat_green_transformed], bin_size=5, bin_range=((0, 1000), (0, 1400))
    )
    # plt.show()

    plt.close("all")


def test_homogeneous_matrix():
    matrix_out = _homogeneous_matrix()
    result = np.identity(4)
    assert np.array_equal(matrix_out, result)

    matrix = ((1, 2, 3), (4, 5, 6), (7, 8, 9))
    offset = (10, 20, 30)
    matrix_out = _homogeneous_matrix(matrix, offset)
    result = np.array(((1, 2, 3, 10), (4, 5, 6, 20), (7, 8, 9, 30), (0, 0, 0, 1)))
    assert np.array_equal(matrix_out, result)

    matrix = ((1, 2), (3, 4))
    offset = (10, 20)
    matrix_out = _homogeneous_matrix(matrix, offset)
    result = np.array(((1, 2, 10), (3, 4, 20), (0, 0, 1)))
    assert np.array_equal(matrix_out, result)


@pytest.mark.parametrize(
    "fixture_name, expected",
    [
        ("locdata_empty", 0),
        ("locdata_single_localization", 4),
        ("locdata_2d", 4),
        ("locdata_3d", 5),
        ("locdata_non_standard_index", 4),
    ],
)
def test_standard_locdata_objects(
    locdata_empty,
    locdata_single_localization,
    locdata_2d,
    locdata_3d,
    locdata_non_standard_index,
    fixture_name,
    expected,
):
    locdata = eval(fixture_name)
    new_locdata = transform_affine(locdata)
    assert len(new_locdata.data.columns) == expected


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
@pytest.mark.parametrize(
    "fixture_name, expected",
    [
        ("locdata_empty", 0),
        ("locdata_single_localization", 4),
        ("locdata_2d", 4),
        ("locdata_3d", 5),
        ("locdata_non_standard_index", 4),
    ],
)
def test_standard_locdata_objects_open3d(
    locdata_empty,
    locdata_single_localization,
    locdata_2d,
    locdata_3d,
    locdata_non_standard_index,
    fixture_name,
    expected,
):
    locdata = eval(fixture_name)
    new_locdata = transform_affine(locdata, method="open3d")
    assert len(new_locdata.data.columns) == expected


def test_transformation_affine_2d(locdata_2d):
    new_locdata = transform_affine(locdata_2d)
    assert np.array_equal(new_locdata.coordinates, locdata_2d.coordinates)
    assert len(new_locdata.data.columns) == 4
    assert new_locdata.meta.history[-1].name == "transform_affine"
    assert (
        "'matrix': None, 'offset': None, 'pre_translation': None, 'method': 'numpy'"
        in new_locdata.meta.history[-1].parameter
    )

    matrix = ((-1, 0), (0, -1))
    offset = (10, 10)
    pre_translation = (100, 100)

    new_locdata = transform_affine(locdata_2d, matrix, offset)
    points_target = ((9, 9), (9, 5), (8, 7), (7, 4), (6, 8), (5, 5))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 4

    new_locdata = transform_affine(
        locdata_2d, offset=offset, pre_translation=pre_translation
    )
    points_target = ((11, 11), (11, 15), (12, 13), (13, 16), (14, 12), (15, 15))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 4


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_transformation_affine_2d_open3d(locdata_2d):
    new_locdata = transform_affine(locdata_2d, method="open3d")
    assert np.array_equal(new_locdata.coordinates, locdata_2d.coordinates)
    assert len(new_locdata.data.columns) == 4

    matrix = ((-1, 0), (0, -1))
    offset = (10, 10)
    pre_translation = (100, 100)

    new_locdata = transform_affine(locdata_2d, matrix, offset, method="open3d")
    points_target = ((9, 9), (9, 5), (8, 7), (7, 4), (6, 8), (5, 5))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 4

    new_locdata = transform_affine(
        locdata_2d, offset=offset, pre_translation=pre_translation, method="open3d"
    )
    points_target = ((11, 11), (11, 15), (12, 13), (13, 16), (14, 12), (15, 15))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 4


def test_transformation_affine_3d(locdata_3d):
    new_locdata = transform_affine(locdata_3d)
    assert np.array_equal(new_locdata.coordinates, locdata_3d.coordinates)
    assert len(new_locdata.data.columns) == 5
    assert new_locdata.meta.history[-1].name == "transform_affine"
    assert (
        "'matrix': None, 'offset': None, 'pre_translation': None, 'method': 'numpy'"
        in new_locdata.meta.history[-1].parameter
    )

    matrix = ((-1, 0, 0), (0, -1, 0), (0, 0, -1))
    offset = (10, 10, 10)
    pre_translation = (100, 100, 100)

    new_locdata = transform_affine(locdata_3d, matrix, offset)
    points_target = ((9, 9, 9), (9, 5, 8), (8, 7, 5), (7, 4, 6), (6, 8, 7), (5, 5, 8))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 5

    new_locdata = transform_affine(
        locdata_3d, offset=offset, pre_translation=pre_translation
    )
    points_target = (
        (11, 11, 11),
        (11, 15, 12),
        (12, 13, 15),
        (13, 16, 14),
        (14, 12, 13),
        (15, 15, 12),
    )
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 5


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_transformation_affine_3d_open3d(locdata_3d):
    new_locdata = transform_affine(locdata_3d, method="open3d")
    assert np.array_equal(new_locdata.coordinates, locdata_3d.coordinates)
    assert len(new_locdata.data.columns) == 5

    matrix = ((-1, 0, 0), (0, -1, 0), (0, 0, -1))
    offset = (10, 10, 10)
    pre_translation = (100, 100, 100)

    new_locdata = transform_affine(locdata_3d, matrix, offset, method="open3d")
    points_target = ((9, 9, 9), (9, 5, 8), (8, 7, 5), (7, 4, 6), (6, 8, 7), (5, 5, 8))
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 5

    new_locdata = transform_affine(
        locdata_3d, offset=offset, pre_translation=pre_translation, method="open3d"
    )
    points_target = (
        (11, 11, 11),
        (11, 15, 12),
        (12, 13, 15),
        (13, 16, 14),
        (14, 12, 13),
        (15, 15, 12),
    )
    assert np.array_equal(new_locdata.coordinates, points_target)
    assert len(new_locdata.data.columns) == 5


def test_overlay(locdata_two_cluster_2d):
    _, clust = cluster_dbscan(locdata_two_cluster_2d, eps=2, min_samples=1)
    new_locdata = overlay(clust.references)
    assert new_locdata.meta.history[0].name == "overlay"
    assert len(new_locdata) == 2
    assert 0 == pytest.approx(new_locdata.coordinates, abs=0.1)

    # various translations
    new_locdata = overlay(clust.references, centers=None)
    assert len(new_locdata) == 2
    assert np.array_equal(new_locdata.coordinates, clust.coordinates)

    centers = [ref.centroid for ref in clust.references]
    new_locdata = overlay(clust.references, centers=centers)
    assert len(new_locdata) == 2
    assert 0 == pytest.approx(new_locdata.coordinates, abs=0.1)

    new_locdata = overlay(clust.references, centers="bb")
    assert len(new_locdata) == 2
    assert 0 == pytest.approx(new_locdata.coordinates, abs=0.1)

    new_locdata = overlay(clust.references, centers="obb")
    assert len(new_locdata) == 2
    assert 0 == pytest.approx(new_locdata.coordinates, abs=0.1)

    new_locdata = overlay(clust.references, centers="ch")
    assert len(new_locdata) == 2
    assert 0 == pytest.approx(new_locdata.coordinates, abs=0.1)

    for reference in clust.references:
        reference.region = reference.bounding_box.region
    new_locdata = overlay(clust.references, centers="region")
    assert len(new_locdata) == 2
    assert 0 == pytest.approx(new_locdata.coordinates, abs=0.1)

    with pytest.raises(ValueError):
        overlay(clust.references, centers="undefined")
    with pytest.raises(ValueError):
        overlay(clust.references, centers=[(1, 1)])

    # various rotations
    orientations = [ref.oriented_bounding_box.angle for ref in clust.references]
    new_locdata = overlay(clust.references, orientations=orientations)
    assert len(new_locdata) == 2
    assert 0 == pytest.approx(new_locdata.coordinates)

    new_locdata = overlay(clust.references, orientations="orientation_obb")
    assert len(new_locdata) == 2
    assert 0 == pytest.approx(new_locdata.coordinates)

    new_locdata = overlay(clust.references, orientations="orientation_im")
    assert len(new_locdata) == 2
    assert 0 == pytest.approx(new_locdata.coordinates)

    with pytest.raises(ValueError):
        overlay(clust.references, orientations="undefined")
    with pytest.raises(ValueError):
        overlay(clust.references, orientations=[(180,)])

    with pytest.raises(TypeError):
        overlay([])
    with pytest.raises(TypeError):
        overlay(clust)



"""

Constants used throughout the project.

.. autosummary::
   :toctree: ./

   PROPERTY_KEYS
   RAPIDSTORM_KEYS
   ELYRA_KEYS
   THUNDERSTORM_KEYS
   NANOIMAGER_KEYS
   SMLM_KEYS
   DECODE_KEYS
   SMAP_KEYS
"""
from __future__ import annotations

from dataclasses import dataclass
from enum import Enum, unique

import matplotlib.colors as mcolors

from locan.dependencies import HAS_DEPENDENCY

if HAS_DEPENDENCY["colorcet"]:
    from colorcet import m_coolwarm, m_fire, m_glasbey_dark, m_gray

__all__ = [
    "PROPERTY_KEYS",
    "PropertyKey",
    "PropertyDescription",
    "HullType",
    "FileType",
    "RenderEngine",
    "ColorMaps",
    "RAPIDSTORM_KEYS",
    "ELYRA_KEYS",
    "THUNDERSTORM_KEYS",
    "NANOIMAGER_KEYS",
    "SMLM_KEYS",
    "DECODE_KEYS",
    "SMAP_KEYS",
]


@dataclass
class PropertyDescription:
    """
    A property of a localization or group of localizations
    representing column names in `LocData.data` and `LocData.properties`.

    Attributes
    ----------
    name:
        property name
    type:
        property type
    unit_SI:
        SI unit that is appropriate for property
    unit:
        The actual unit currently used
    description:
        Explanation what the property represents
    """

    name: str
    type: str
    unit_SI: str | None = None
    unit: str | None = None
    description: str = ""


@unique
class PropertyKey(Enum):
    """
    Property descriptions for standard properties used in `locan.LocData` and throughout locan.
    """

    index = PropertyDescription(
        "index",
        "integer",
        description="a positive integer identifying the localization",
    )
    original_index = PropertyDescription("original_index", "integer")
    position_x = PropertyDescription(
        "position_x", "float", unit_SI="m", description="spatial coordinate."
    )
    position_y = PropertyDescription(
        "position_y", "float", unit_SI="m", description="spatial coordinate."
    )
    position_z = PropertyDescription(
        "position_z", "float", unit_SI="m", description="spatial coordinate."
    )
    frame = PropertyDescription(
        "frame", "integer", description="frame  number in which the localization occurs"
    )
    frames_number = PropertyDescription(
        "frames_number",
        "integer",
        description="number of frames that contribute to a merged localization",
    )
    frames_missing = PropertyDescription(
        "frames_missing",
        "integer",
        description="number of frames that occurred between two successive localizations",
    )
    time = PropertyDescription("time", "float", unit_SI="s")
    intensity = PropertyDescription(
        "intensity",
        "float",
        description="intensity or emission strength as estimated by the fitter",
    )
    local_background = PropertyDescription(
        "local_background",
        "float",
        description="background in the neighborhood of localization as estimated by the fitter",
    )
    local_background_sigma = PropertyDescription(
        "local_background_sigma",
        "float",
        description="variation of local background in terms of standard deviation",
    )
    signal_noise_ratio = PropertyDescription(
        "signal_noise_ratio",
        "float",
        description="ratio between mean intensity (i.e. intensity for a single localization) "
        "and the standard deviation of local_background "
        "(i.e. local_background_sigma for a single localization)",
    )
    signal_background_ratio = PropertyDescription(
        "signal_background_ratio",
        "float",
        description="ratio between mean intensity (i.e. intensity for a single localization) and the local_background",
    )
    chi_square = PropertyDescription(
        "chi_square",
        "float",
        description="chi-square value of the fitting procedure as estimated by the fitter",
    )
    two_kernel_improvement = PropertyDescription(
        "two_kernel_improvement",
        "float",
        description="a rapidSTORM parameter describing the improvement from two kernel fitting",
    )
    psf_amplitude = PropertyDescription("psf_amplitude", "float")
    psf_width = PropertyDescription(
        "psf_width",
        "float",
        description="full-width-half-max of the fitted Gauss-function - "
        "being isotropic or representing the root-mean-square of psf_width_c for all dimensions",
    )
    psf_width_x = PropertyDescription(
        "psf_width_x",
        "float",
        description="full-width-half-max of the fitted Gauss-function in x-dimension as estimated by the fitter",
    )
    psf_width_y = PropertyDescription(
        "psf_width_y",
        "float",
        description="full-width-half-max of the fitted Gauss-function in y-dimension as estimated by the fitter",
    )
    psf_width_z = PropertyDescription(
        "psf_width_z",
        "float",
        description="full-width-half-max of the fitted Gauss-function in z-dimension as estimated by the fitter",
    )
    psf_half_width = PropertyDescription("psf_half_width", "float")
    psf_half_width_x = PropertyDescription("psf_half_width_x", "float")
    psf_half_width_y = PropertyDescription("psf_half_width_y", "float")
    psf_half_width_z = PropertyDescription("psf_half_width_z", "float")
    psf_sigma = PropertyDescription(
        "psf_sigma",
        "float",
        description="sigma of the fitted Gauss-function - "
        "being isotropic or representing the root-mean-square of psf_sigma_c for all dimensions",
    )
    psf_sigma_x = PropertyDescription(
        "psf_sigma_x",
        "float",
        description="sigma of the fitted Gauss-function in x-dimension as estimated by the fitter",
    )
    psf_sigma_y = PropertyDescription(
        "psf_sigma_y",
        "float",
        description="sigma of the fitted Gauss-function in y-dimension as estimated by the fitter",
    )
    psf_sigma_z = PropertyDescription(
        "psf_sigma_z",
        "float",
        description="sigma of the fitted Gauss-function in z-dimension as estimated by the fitter",
    )
    uncertainty = PropertyDescription(
        "uncertainty",
        "float",
        description="localization error for all dimensions or representing a value proportional to "
        "psf_sigma / sqrt(intensity) or "
        "representing the root-mean-square of uncertainty_c for all dimensions.",
    )
    uncertainty_x = PropertyDescription(
        "uncertainty_x",
        "float",
        description="localization error in x-dimension estimated by a fitter "
        "or representing a value proportional to psf_sigma_x / sqrt(intensity)",
    )
    uncertainty_y = PropertyDescription(
        "uncertainty_y",
        "float",
        description="localization error in y-dimension estimated by a fitter "
        "or representing a value proportional to psf_sigma_y / sqrt(intensity)",
    )
    uncertainty_z = PropertyDescription(
        "uncertainty_z",
        "float",
        description="localization error in z-dimension estimated by a fitter "
        "or representing a value proportional to psf_sigma_z / sqrt(intensity)",
    )
    channel = PropertyDescription(
        "channel", "integer", description="identifier for an imaging channel"
    )
    slice_z = PropertyDescription(
        "slice_z", "float", description="identifier for an imaging slice"
    )
    plane = PropertyDescription(
        "plane", "integer", description="identifier for an imaging plane"
    )
    cluster_label = PropertyDescription(
        "cluster_label", "integer", description="identifier for a localization cluster"
    )

    @classmethod
    def coordinate_labels(cls) -> list[PropertyKey]:
        """Property descriptions for properties representing spatial coordinates"""
        return [cls.position_x, cls.position_y, cls.position_z]

    @classmethod
    def summary(cls) -> str:
        """A formatted string representation of PropertyKey showing all elements with attributes."""
        lines = []
        for element in cls:
            lines.append(f"{element.value.name}")
            lines.append(
                f"type: {element.value.type}, unit_SI: {element.value.unit_SI}, unit: {element.value.unit}"
            )
            lines.append(f"{element.value.description}\n")
        return "\n".join(lines)


#: Keys for the most common LocData properties.
#: Values suggest a type for conversion.
#: If 'integer', 'signed', 'unsigned', 'float' :func:`pandas.to_numeric` can be applied.
#: Otherwise :func:`pandas.astype` can be applied.
PROPERTY_KEYS = {
    item.name: item.value.type for item in PropertyKey if item.value.type is not None
}


class HullType(Enum):
    """
    Hull definitions that are supported for `LocData` objects.
    """

    BOUNDING_BOX = "bounding_box"
    CONVEX_HULL = "convex_hull"
    ORIENTED_BOUNDING_BOX = "oriented_bounding_box"
    ALPHA_SHAPE = "alpha_shape"


# File types
class FileType(Enum):
    """
    File types for localization files.

    The listed file types are supported with input/output functions in :func:`io.io_locdata`.
    The types correspond to the metadata keys for LocData objects. That is they are equal to the file types in
    the protobuf message `locan.data.metadata_pb2.Metadata`.
    """

    UNKNOWN_FILE_TYPE = 0
    CUSTOM = 1
    RAPIDSTORM = 2
    ELYRA = 3
    THUNDERSTORM = 4
    ASDF = 5
    NANOIMAGER = 6
    RAPIDSTORMTRACK = 7
    SMLM = 8
    DECODE = 9
    SMAP = 10


class RenderEngine(Enum):
    """
    Engine to be use for rendering and displaying localization data as 2d or 3d images.

    Each engine represents a library to be used as backend for rendering and plotting.
    """

    if not HAS_DEPENDENCY["mpl_scatter_density"]:
        _ignore_ = "MPL_SCATTER_DENSITY"
    if not HAS_DEPENDENCY["napari"]:
        _ignore_ = "NAPARI"
    MPL = 0
    """matplotlib"""
    MPL_SCATTER_DENSITY = 1
    """mpl-scatter-density"""
    NAPARI = 2
    """napari"""


class ColorMaps(Enum):
    """
    Preferred colormaps to be used for visualization.
    """

    if HAS_DEPENDENCY["colorcet"]:
        COLORMAP_CONTINUOUS = m_fire
        COLORMAP_CONTINUOUS_GRAY = m_gray
        COLORMAP_DIVERGING = m_coolwarm
        COLORMAP_CATEGORICAL = m_glasbey_dark
    else:
        COLORMAP_CONTINUOUS = mcolors.Colormap("viridis")
        COLORMAP_DIVERGING = mcolors.Colormap("coolwarm")
        COLORMAP_CATEGORICAL = mcolors.Colormap("tab20")


#: Mapping column names in RapidSTORM files to `LocData` property keys
RAPIDSTORM_KEYS = {
    "Position-0-0": "position_x",
    "Position-1-0": "position_y",
    "Position-2-0": "position_z",
    "ImageNumber-0-0": "frame",
    "Amplitude-0-0": "intensity",
    "FitResidues-0-0": "chi_square",
    "LocalBackground-0-0": "local_background",
    "TwoKernelImprovement-0-0": "two_kernel_improvement",
    "Position-0-0-uncertainty": "uncertainty_x",
    "Position-1-0-uncertainty": "uncertainty_y",
    "Position-2-0-uncertainty": "uncertainty_z",
}


#: Mapping column names in Zeiss Elyra files to `LocData` property keys
ELYRA_KEYS = {
    "Index": "original_index",
    "First Frame": "frame",
    "Number Frames": "frames_number",
    "Frames Missing": "frames_missing",
    "Position X [nm]": "position_x",
    "Position Y [nm]": "position_y",
    "Position Z [nm]": "position_z",
    "Precision [nm]": "uncertainty",
    "Number Photons": "intensity",
    "Background variance": "local_background_sigma",
    "Chi square": "chi_square",
    "PSF half width [nm]": "psf_half_width",
    "PSF width [nm]": "psf_width",
    "Channel": "channel",
    "Z Slice": "slice_z",
}


#: Mapping column names in Thunderstorm files to `LocData` property keys
THUNDERSTORM_KEYS = {
    "id": "original_index",
    "frame": "frame",
    "x [nm]": "position_x",
    "y [nm]": "position_y",
    "z [nm]": "position_z",
    "uncertainty [nm]": "uncertainty",
    "uncertainty_xy [nm]": "uncertainty_x",
    "uncertainty_z [nm]": "uncertainty_z",
    "intensity [photon]": "intensity",
    "offset [photon]": "local_background",
    "bkgstd [photon]": "local_background_sigma",
    "chi2": "chi_square",
    "sigma1 [nm]": "psf_sigma_x",
    "sigma2 [nm]": "psf_sigma_y",
    "sigma [nm]": "psf_sigma",
    "detections": "frames_number",
}


#: Mapping column names in Nanoimager files to `LocData` property keys
NANOIMAGER_KEYS = {
    "Channel": "channel",
    "Frame": "frame",
    "X (nm)": "position_x",
    "Y (nm)": "position_y",
    "Z (nm)": "position_z",
    "Photons": "intensity",
    "Background": "local_background",
}


#: Mapping column names in SMLM files to `LocData` property keys
SMLM_KEYS = {
    "id": "original_index",
    "frame": "frame",
    "x": "position_x",
    "y": "position_y",
    "z": "position_z",
    "x_position": "position_x",
    "y_position": "position_y",
    "z_position": "position_z",
    "uncertainty [nm]": "uncertainty",
    "uncertainty_xy [nm]": "uncertainty_x",
    "uncertainty_z [nm]": "uncertainty_z",
    "intensity": "intensity",
    "Amplitude_0_0": "intensity",
    "background": "local_background",
    "LocalBackground_0_0": "local_background",
    "FitResidues_0_0": "chi_square",
}

#: Mapping column names in DECODE files to `LocData` property keys
DECODE_KEYS = {
    "id": "original_index",
    "frame_ix": "frame",
    "x": "position_x",
    "y": "position_y",
    "z": "position_z",
    "bg": "local_background",
    "phot": "intensity",
}

#: Mapping column names in SMAP files to `LocData` property keys
SMAP_KEYS = {
    "frame": "frame",
    "xnm": "position_x",
    "ynm": "position_y",
    "znm": "position_z",
    "bg": "local_background",
    "phot": "intensity",
    "channel": "channel",
    "xnmerr": "uncertainty_x",
    "ynmerr": "uncertainty_y",
    "znmerr": "uncertainty_z",
}



import os
from pathlib import Path

import pytest
import tifffile as tif

from locan import ROOT_DIR, load_locdata
from locan.dependencies import HAS_DEPENDENCY
from locan.scripts.script_check import render_locs_per_frame_napari, sc_check

if HAS_DEPENDENCY["napari"]:
    import napari


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_render_locs_per_frame_napari(locdata_2d):
    images_path = Path(ROOT_DIR / "tests/test_data/images.tif")
    assert images_path.exists()
    locdata_path = Path(ROOT_DIR / "tests/test_data/rapidSTORM_from_images.txt")
    assert locdata_path.exists()

    image_stack = tif.imread(str(images_path))
    locdata = load_locdata(locdata_path, file_type=2)

    viewer = render_locs_per_frame_napari(
        images=image_stack,
        pixel_size=133,
        locdata=locdata,
        viewer=None,
        transpose=True,
        kwargs_image=None,
        kwargs_points=None,
    )
    assert viewer
    napari.run()


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_script_check():
    images_path = Path(ROOT_DIR / "tests/test_data/images.tif")
    assert images_path.exists()
    locdata_path = Path(ROOT_DIR / "tests/test_data/rapidSTORM_from_images.txt")
    assert locdata_path.exists()
    sc_check(
        pixel_size=133, file_images=images_path, file_locdata=locdata_path, file_type=2
    )


@pytest.mark.gui
@pytest.mark.skipif(not HAS_DEPENDENCY["napari"], reason="Test requires napari.")
def test_script_check_from_sys(capfd):
    images_path = Path(ROOT_DIR / "tests/test_data/images.tif")
    assert images_path.exists()
    locdata_path = Path(ROOT_DIR / "tests/test_data/rapidSTORM_from_images.txt")
    assert locdata_path.exists()
    exit_status = os.system(
        f"locan check 133 -f {str(images_path)} -l {str(locdata_path)} -t 2"
    )
    captured = capfd.readouterr()
    print(captured.out)
    assert exit_status == 0



"""

Filter localization data.

This module provides functions for filtering LocData objects.
The functions take LocData as input and compute new LocData objects.

"""
from __future__ import annotations

import sys

import numpy as np
from sklearn.neighbors import NearestNeighbors

from locan.configuration import N_JOBS
from locan.constants import HullType
from locan.data.locdata import LocData
from locan.data.metadata_utils import _modify_meta
from locan.data.region import Region2D, RoiRegion

__all__ = [
    "select_by_condition",
    "select_by_region",
    "select_by_image_mask",
    "exclude_sparse_points",
    "random_subset",
    "localizations_in_cluster_regions",
]


def select_by_condition(locdata, condition) -> LocData:
    """
    Select by specifying conditions on data properties.

    Parameters
    ----------
    locdata : LocData
        Specifying the localization data from which to select.
    condition : str
        Conditions as input in select method.
        More precise: query specifications to be used with pandas query.

    Returns
    -------
    LocData
        A new instance of LocData referring to the specified dataset.
    """
    local_parameter = locals()

    # select
    new_indices = locdata.data.query(condition).index.values.tolist()

    # instantiate
    new_locdata = LocData.from_selection(locdata=locdata, indices=new_indices)

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=None,
    )
    new_locdata.meta = meta_

    return new_locdata


def select_by_region(locdata, region, loc_properties=None, reduce=True) -> LocData:
    """
    Select localizations from `locdata` that are within `region` and return
    a new LocData object.
    Selection is with respect to loc_properties or to localization coordinates
    that correspond to region dimension.

    Parameters
    ----------
    locdata : LocData
        Localization data that is tested for being inside the region.
    region : Region
        Tested region
    loc_properties : list[str] | None
        Localization properties to be tested.
    reduce : bool
        Return the reduced LocData object or keep references alive.

    Returns
    -------
    LocData
        A new instance of LocData with all localizations within region of
        interest.

    Note
    ----
    Points on boundary of regions are considered outside if region is a
    shapely object, but inside if region is a
    matplotlib (which is the case for RoiRegion) object.
    """
    local_parameter = locals()

    if loc_properties is None:
        loc_properties_ = locdata.coordinate_labels[0 : region.dimension]
    else:
        loc_properties_ = loc_properties

    points = locdata.data[list(loc_properties_)].values

    if isinstance(region, (Region2D, RoiRegion)):
        indices_inside = region.contains(points)
        locdata_indices_to_keep = locdata.data.index[indices_inside]
        new_locdata = LocData.from_selection(
            locdata=locdata, indices=locdata_indices_to_keep
        )
        new_locdata.region = region
    else:
        raise NotImplementedError("Only Region2D has been implemented.")

    # finish
    if reduce:
        new_locdata.reduce()

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=None,
    )
    new_locdata.meta = meta_

    return new_locdata


# todo: implement
def select_by_image_mask(selection, mask, bins):
    """
    Select by masking using a binary image(e.g. generated by thresholding a
    transmitted-light microscopy image.

    Parameters
    ----------
    selection : Selection
        specifying the localization data from which to select.
    mask : Sequence
        binary image.
    bins : Bins
        Bins in units of localization coordinates.

    Returns
    -------
    Selection
        a new instance of Selection referring to the specified dataset.
    """
    raise NotImplementedError


def exclude_sparse_points(
    locdata, other_locdata=None, radius=50, min_samples=5
) -> LocData:
    """
    Exclude localizations by thresholding a local density.

    A subset of localizations, that exhibit a small local density of
    localizations from locdata or alternatively from other_locdata,
    is identified as noise and excluded.
    Noise is identified by using a nearest-neighbor search
    (:class:`sklearn.neighbors.NearestNeighbors`) to find all
    localizations within a circle (sphere) of the given `radius`.
    If the number of localizations is below the
    threshold value `min_samples`, the localization is considered to be noise.

    The method identifies the same noise points as done by the clustering
    algorithm DBSCAN [1]_.

    Parameters
    ----------
    locdata : LocData
        Specifying the localization data from which to exclude localization
        data.
    other_locdata : LocData
        Specifying the localization data on which to compute local density.
    radius: float
        Radius of a circle or sphere in which neighbors are identified
        (equivalent to epsilon in DBSCAN).
    min_samples : int
        The minimum number of samples in the neighborhood that need to be
        found for each localization to not be
        identified as noise (equivalent to minPoints in DBSCAN).

    Returns
    -------
    LocData
        All localizations except those identified as sparse (noise) points.

    References
    ----------
    .. [1] Martin Ester, Hans-Peter Kriegel, Jörg Sander, Xiaowei Xu,
       A density-based algorithm for discovering clusters in large spatial
       databases with noise.
       In: Evangelos Simoudis, Jiawei Han, Usama M. Fayyad (Hrsg.):
       Proceedings of the Second International Conference
       on Knowledge Discovery and Data Mining (KDD-96). AAAI Press, 1996,
       S. 226-231, ISBN 1-57735-004-9.
    """
    local_parameter = locals()

    if other_locdata is None:
        nn = NearestNeighbors(metric="euclidean", n_jobs=N_JOBS).fit(
            locdata.coordinates
        )
        neighbor_points_list = nn.radius_neighbors(radius=radius, return_distance=False)
        # if points is not provided the query point is not considered its own neighbor.
    else:
        nn = NearestNeighbors(metric="euclidean", n_jobs=N_JOBS).fit(
            other_locdata.coordinates
        )
        neighbor_points_list = nn.radius_neighbors(
            locdata.coordinates, radius=radius, return_distance=False
        )

    indices_to_keep = [len(pts) >= min_samples for pts in neighbor_points_list]
    locdata_indices_to_keep = locdata.data.index[indices_to_keep]
    new_locdata = LocData.from_selection(locdata, locdata_indices_to_keep)

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=None,
    )
    new_locdata.meta = meta_

    return new_locdata


def random_subset(locdata, n_points, replace=True, seed=None) -> LocData:
    """
    Take a random subset of localizations.

    Parameters
    ----------
    locdata : LocData
        Specifying the localization data from which to select localization data.
    n_points : int
        Number of localizations to randomly choose from locdata.
    replace : bool
        Indicate if sampling is with or without replacement
    seed : None, int, array_like[ints], numpy.random.SeedSequence, numpy.random.BitGenerator, numpy.random.Generator
        Random number generation seed

    Returns
    -------
    LocData
        A new instance of LocData carrying the subset of localizations.
    """
    local_parameter = locals()

    if not locdata:
        return locdata

    rng = np.random.default_rng(seed)

    indices = rng.choice(locdata.data.index, size=n_points, replace=replace)
    new_locdata = LocData.from_selection(locdata, indices)

    # update metadata
    meta_ = _modify_meta(
        locdata,
        new_locdata,
        function_name=sys._getframe().f_code.co_name,
        parameter=local_parameter,
        meta=None,
    )
    new_locdata = LocData.from_selection(locdata, indices, meta=meta_)

    return new_locdata


def localizations_in_cluster_regions(
    locdata, collection, hull_type=HullType.CONVEX_HULL
):
    """
    Identify localizations from `locdata` within the regions of all
    `collection` elements.

    Parameters
    ----------
    locdata : LocData
        Localization data that is tested for being inside the region

    collection : LocData | list[LocData]
        A set of Locdata objects collected in a collection or list.

    hull_type : HullType | str
        The hull type for each LocData object that is used to define the
        region.

    Returns
    --------
    LocData
        A collection of LocData objects with all elements of locdata
        contained by the region.
    """
    locdatas = []
    if isinstance(hull_type, str):
        hull_type = HullType[hull_type.upper()].value
    else:
        hull_type = hull_type.value

    if isinstance(collection, LocData):
        if isinstance(collection.references, list):  # this case covers pure collections
            for ref in collection.references:
                cregion = getattr(ref, hull_type).region
                locdata_selection = select_by_region(locdata=locdata, region=cregion)
                locdatas.append(locdata_selection)
        else:  # this case covers selections of collections
            for index in collection.indices:
                cregion = getattr(
                    collection.references.references[index], hull_type
                ).region
                locdata_selection = select_by_region(locdata=locdata, region=cregion)
                locdatas.append(locdata_selection)
    else:  # this case covers list of LocData objects
        for ref in collection:
            cregion = getattr(ref, hull_type).region
            locdata_selection = select_by_region(locdata=locdata, region=cregion)
            locdatas.append(locdata_selection)

    new_collection = LocData.from_collection(locdatas)

    return new_collection



import copy

import numpy as np
import pandas as pd
import pytest

from locan import AlphaShape, LocData, Rectangle, Region
from locan.data import metadata_pb2

# fixtures for DataFrames (fixtures for LocData are defined in conftest.py)


@pytest.fixture()
def df_simple():
    dict_ = {"position_x": [0, 0, 1, 4, 5], "position_y": [0, 1, 3, 4, 1]}
    return pd.DataFrame.from_dict(dict_)


@pytest.fixture()
def df_line():
    dict_ = {"position_x": [1, 2, 3, 4, 5], "position_y": [1, 2, 3, 4, 5]}
    return pd.DataFrame.from_dict(dict_)


@pytest.fixture()
def df_empty():
    dict_ = {}
    return pd.DataFrame.from_dict(dict_)


@pytest.fixture()
def df_other_simple():
    dict_ = {"position_x": [0, 0, 1, 4, 5], "position_y": [10, 11, 13, 14, 11]}
    return pd.DataFrame.from_dict(dict_)


# tests

COMMENT_METADATA = metadata_pb2.Metadata(comment="some user comment")


def test_LocData(df_simple, caplog):
    dat = LocData(dataframe=df_simple, meta=COMMENT_METADATA)
    assert len(dat) == 5
    assert dat.coordinate_labels == ["position_x", "position_y"]
    assert dat.dimension == 2
    assert np.array_equal(dat.centroid, [2.0, 1.8])
    assert dat.centroid[0] == dat.properties["position_x"]
    for x, y in zip(dat.coordinates, [[0, 0], [0, 1], [1, 3], [4, 4], [5, 1]]):
        assert np.all(x == np.array(y))
    assert dat.meta.comment == COMMENT_METADATA.comment
    # The following test runs ok when testing this function alone.
    # assert dat.meta.identifier == '1'
    assert dat.bounding_box.region_measure == 20
    assert "region_measure_bb" in dat.properties
    assert "localization_density_bb" in dat.properties
    assert dat.convex_hull.region_measure == 12.5
    assert "region_measure_ch" in dat.properties
    assert "localization_density_ch" in dat.properties
    assert round(dat.oriented_bounding_box.region_measure) == 16
    assert "region_measure_obb" in dat.properties
    assert "localization_density_obb" in dat.properties
    assert dat.region is None
    dat.region = Rectangle()
    assert caplog.record_tuples == [
        ("locan.data.locdata", 30, "Not all coordinates are within region.")
    ]
    dat.region = dat.bounding_box.region
    assert repr(dat.region) == repr(dat.bounding_box.region)
    assert isinstance(dat.update_alpha_shape(alpha=1).alpha_shape, AlphaShape)
    assert "region_measure_as" in dat.properties
    assert "localization_density_as" in dat.properties

    assert len(dat.inertia_moments.eigenvalues) == 2
    assert "orientation_im" in dat.properties
    assert "circularity_im" in dat.properties


def test_LocData_empty(df_empty):
    dat = LocData()
    assert dat.data.empty
    assert dat.properties == {"localization_count": 0, "region_measure_bb": 0}
    assert len(dat) == 0
    assert dat.coordinate_labels == []
    assert dat.dimension == 0

    dat = LocData(dataframe=df_empty)
    assert dat.data.empty
    assert dat.properties == {"localization_count": 0, "region_measure_bb": 0}
    assert len(dat) == 0
    assert dat.coordinate_labels == []
    assert dat.dimension == 0
    # hulls are tested with locdata fixtures further down.


def test_LocData_from_dataframe(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    # print(dat.properties.keys())
    assert dat.bounding_box is not None
    assert list(dat.properties.keys()) == [
        "localization_count",
        "position_x",
        "position_y",
        "region_measure_bb",
        "localization_density_bb",
        "subregion_measure_bb",
    ]
    assert len(dat) == 5
    assert dat.meta.comment == COMMENT_METADATA.comment


def test_LocData_from_dataframe_empty(df_empty):
    dat = LocData.from_dataframe(dataframe=df_empty)
    assert len(dat) == 0
    assert dat.coordinate_labels == []
    # print(dat.data)


def test_LocData_from_dataframe_with_meta_dict(df_simple):
    dat = LocData.from_dataframe(
        dataframe=df_simple, meta={"comment": "some user comment"}
    )
    assert len(dat) == 5
    assert dat.references is None
    assert dat.meta.comment == COMMENT_METADATA.comment


def test_LocData_from_coordinates():
    coordinates = [(200, 500), (200, 600), (900, 650), (1000, 600)]
    dat = LocData.from_coordinates(coordinates=coordinates, meta=COMMENT_METADATA)
    assert np.array_equal(dat.coordinates, np.asarray(coordinates))
    assert all(item in ["position_x", "position_y"] for item in dat.coordinate_labels)
    assert len(dat) == 4
    assert dat.meta.comment == COMMENT_METADATA.comment

    coordinates = np.array([(200, 500), (200, 600), (900, 650), (1000, 600)])
    dat = LocData.from_coordinates(coordinates=coordinates, meta=COMMENT_METADATA)
    assert np.array_equal(dat.coordinates, np.asarray(coordinates))
    assert all(item in ["position_x", "position_y"] for item in dat.coordinate_labels)
    assert len(dat) == 4

    dat = LocData.from_coordinates(
        coordinates=coordinates,
        coordinate_labels=["position_x", "position_z"],
        meta=dict(comment="special order"),
    )
    assert np.array_equal(dat.coordinates, np.asarray(coordinates))
    assert all(item in ["position_x", "position_z"] for item in dat.coordinate_labels)
    assert len(dat) == 4

    with pytest.raises(ValueError):
        LocData.from_coordinates(
            coordinates=coordinates, coordinate_labels=["position_x", "whatever"]
        )

    dat = LocData.from_coordinates(coordinates=[])
    assert len(dat) == 0


def test_LocData_from_selection(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple)

    sel = LocData.from_selection(locdata=dat, indices=slice(1, 3))
    assert all(sel.data.index == [1, 2, 3])

    sel = LocData.from_selection(locdata=dat)
    assert len(sel) == len(df_simple)

    sel = LocData.from_selection(locdata=dat, indices=slice(4, 10))
    assert all(sel.data.index == [4])

    sel = LocData.from_selection(locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA)
    assert len(sel) == 3
    assert sel.references is dat
    assert sel.meta.comment == COMMENT_METADATA.comment
    assert dat.bounding_box.region_measure != sel.bounding_box.region_measure
    with pytest.raises(AttributeError):
        sel.data = df_simple

    sel_sel = LocData.from_selection(
        locdata=sel, indices=[1, 3], meta={"comment": "Selection of a selection."}
    )
    assert sel_sel.meta.comment == "Selection of a selection."
    assert len(sel_sel) == 2
    assert sel_sel.references is sel

    sel_empty = LocData.from_selection(
        locdata=dat, indices=[], meta={"comment": "This selection is empty."}
    )
    assert sel_empty.data.empty
    assert len(sel_empty) == 0

    dat = LocData()
    sel_empty = LocData.from_selection(locdata=dat, indices=[])
    assert sel_empty.data.empty
    assert len(sel_empty) == 0

    sel_empty = LocData.from_selection(locdata=dat, indices=[1, 3])
    assert sel_empty.data.empty
    assert len(sel_empty) == 0


def test_LocData_from_collection(df_simple):
    col = LocData.from_collection([LocData(), LocData()], meta=COMMENT_METADATA)
    assert col.dimension == 0
    assert len(col) == 2
    assert col.properties == {"localization_count": 2, "region_measure_bb": 0}
    assert col.meta.comment == COMMENT_METADATA.comment

    dat = LocData.from_dataframe(dataframe=df_simple)
    sel_1 = LocData.from_selection(locdata=dat, indices=[0, 1, 2])
    sel_2 = LocData.from_selection(locdata=dat, indices=[3, 4])
    col = LocData.from_collection(
        [sel_1, sel_2], meta=dict(comment="yet another comment")
    )
    assert len(col.references) == 2
    assert len(col) == 2
    assert col.meta.comment == "yet another comment"
    assert set(col.data.columns) == {
        "localization_count",
        "localization_density_bb",
        "position_x",
        "position_y",
        "region_measure_bb",
        "subregion_measure_bb",
    }

    with pytest.warns(UserWarning):
        col.update_convex_hulls_in_references()
    assert set(col.data.columns) == {
        "localization_count",
        "localization_density_bb",
        "position_x",
        "position_y",
        "region_measure_bb",
        "subregion_measure_bb",
        "region_measure_ch",
        "localization_density_ch",
    }

    col.update_oriented_bounding_box_in_references()
    assert set(col.data.columns) == {
        "circularity_obb",
        "localization_count",
        "localization_density_bb",
        "localization_density_ch",
        "localization_density_obb",
        "orientation_obb",
        "position_x",
        "position_y",
        "region_measure_bb",
        "region_measure_ch",
        "region_measure_obb",
        "subregion_measure_bb",
    }
    # print(col.properties)

    with pytest.warns(UserWarning):
        col.update_alpha_shape_in_references(alpha=1)
    assert set(col.data.columns) == {
        "circularity_obb",
        "localization_count",
        "localization_density_bb",
        "localization_density_ch",
        "localization_density_obb",
        "orientation_obb",
        "position_x",
        "position_y",
        "region_measure_bb",
        "region_measure_ch",
        "region_measure_obb",
        "subregion_measure_bb",
        "region_measure_as",
        "localization_density_as",
    }

    col.update_inertia_moments_in_references()
    assert set(col.data.columns) == {
        "circularity_obb",
        "localization_count",
        "localization_density_bb",
        "localization_density_ch",
        "localization_density_obb",
        "orientation_obb",
        "position_x",
        "position_y",
        "region_measure_bb",
        "region_measure_ch",
        "region_measure_obb",
        "subregion_measure_bb",
        "region_measure_as",
        "localization_density_as",
        "circularity_im",
        "orientation_im",
    }


def test_LocData_selection_from_collection(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple)
    sel = []
    col = LocData.from_collection(sel)
    assert len(col) == 0

    for i in range(4):
        sel.append(LocData.from_selection(locdata=dat, indices=[i]))
    col = LocData.from_collection(sel, meta=COMMENT_METADATA)
    assert len(col) == 4
    assert len(col.references) == 4
    # print(col.data)

    col_sel = LocData.from_selection(locdata=col, indices=[0, 2, 3])
    assert len(col_sel) == 3
    assert col_sel.references is col
    # print(col_sel.data)

    col_sel_sel = LocData.from_selection(locdata=col_sel, indices=[2])
    assert len(col_sel_sel) == 1
    assert col_sel_sel.references is col_sel
    # print(col_sel_sel.data)


def test_LocData_concat(df_simple):
    col = LocData.concat([LocData(), LocData()], meta=COMMENT_METADATA)
    assert col.dimension == 0
    assert len(col) == 0
    assert col.properties == {"localization_count": 0, "region_measure_bb": 0}
    assert col.meta.comment == COMMENT_METADATA.comment

    dat = LocData.from_dataframe(dataframe=df_simple)
    sel_1 = LocData.from_selection(locdata=dat, indices=[0, 1, 2])
    sel_2 = LocData.from_selection(locdata=dat, indices=[3, 4])
    col = LocData.concat([sel_1, sel_2], meta=dict(comment="yet another comment"))
    assert len(col) == 5
    assert col.meta.comment == "yet another comment"
    assert len(col.references) == 2
    sel_1.reduce()
    col = LocData.concat([sel_1, sel_2], meta=COMMENT_METADATA)
    assert len(col.references) == 2
    sel_2.reduce()
    col = LocData.concat([sel_1, sel_2], meta=COMMENT_METADATA)
    assert col.references is None


def test_LocData_reduce(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    sel_1 = LocData.from_selection(
        locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA
    )
    sel_2 = LocData.from_selection(
        locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA
    )
    sel_1.reduce()
    assert sel_1.references is None
    assert len(sel_1) == 3
    assert len(sel_1.data) == len(sel_2.data)
    col = LocData.from_collection([sel_1, sel_2], meta=COMMENT_METADATA)
    col.reduce()
    assert len(col) == 2
    assert col.references is None


@pytest.mark.parametrize(
    "fixture_name, expected",
    [("locdata_empty", 0), ("locdata_single_localization", 1)],
)
def test_locdata_reduce_empty(
    locdata_empty, locdata_single_localization, fixture_name, expected
):
    dat = eval(fixture_name)
    new_dat = dat.reduce()
    assert len(new_dat) == expected


def test_LocData_reset(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    assert dat.properties["position_x"] == 2
    assert dat.properties["region_measure_bb"] == 20
    assert dat.meta.comment == COMMENT_METADATA.comment

    dat.dataframe["position_x"] = [0, 1, 4, 0, 0]
    dat.dataframe["position_y"] = [1, 3, 4, 0, 0]
    dat.reset()
    assert dat.properties["position_x"] == 1
    assert dat.properties["region_measure_bb"] == 16
    assert dat.meta.comment == COMMENT_METADATA.comment


@pytest.mark.parametrize(
    "fixture_name, expected",
    [("locdata_empty", 0), ("locdata_single_localization", 1)],
)
def test_locdata_reset_empty(
    locdata_empty, locdata_single_localization, fixture_name, expected
):
    dat = eval(fixture_name)
    new_dat = dat.reset()
    assert len(new_dat) == expected


def test_LocData_update(df_simple, caplog):
    new_dataframe = pd.DataFrame.from_dict(
        {
            "position_x": [10, 0, 1, 4],
            "position_y": [10, 1, 3, 4],
            "frame": [0, 1, 1, 4],
        }
    )

    dat = LocData()
    dat.update(new_dataframe, meta=COMMENT_METADATA)
    pd.testing.assert_frame_equal(dat.data, new_dataframe)
    assert dat.properties["position_x"] == 3.75
    assert dat.meta.element_count == 4
    assert dat.meta.history[-1].name == "LocData.update"

    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    assert dat.properties["position_x"] == 2
    assert dat.meta.element_count == 5

    dat.update(new_dataframe, meta=dict(comment="yet another comment"))
    pd.testing.assert_frame_equal(dat.data, new_dataframe)
    assert dat.properties["position_x"] == 3.75
    assert dat.meta.element_count == 4
    assert dat.meta.history[-1].name == "LocData.update"

    sel = LocData.from_selection(locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA)
    sel.update(new_dataframe, reset_index=True)
    assert caplog.record_tuples.pop() == (
        "locan.data.locdata",
        30,
        "LocData.reduce() was applied since self.references was not None.",
    )
    pd.testing.assert_frame_equal(sel.data, new_dataframe)
    assert sel.properties["position_x"] == 3.75
    assert sel.meta.element_count == 4
    assert sel.meta.history[-1].name == "LocData.update"

    sel = LocData.from_selection(
        locdata=dat, indices=[1, 2, 3, 4], meta=COMMENT_METADATA
    )
    sel.update(new_dataframe, reset_index=True)
    assert caplog.record_tuples.pop() == (
        "locan.data.locdata",
        30,
        "LocData.reduce() was applied since self.references was not None.",
    )
    pd.testing.assert_frame_equal(sel.data, new_dataframe)
    assert sel.properties["position_x"] == 3.75
    assert sel.meta.element_count == 4
    assert sel.meta.history[-1].name == "LocData.update"


def test_LocData_projection(locdata_3d):
    locdata = copy.deepcopy(locdata_3d)

    # test with single coordinate_label
    new_coordinate_labels = "position_x"
    new_locdata = locdata.projection(coordinate_labels=new_coordinate_labels)

    assert all(
        label in new_coordinate_labels for label in new_locdata.coordinate_labels
    )
    assert new_locdata.dimension == 1
    assert len(new_locdata) == len(locdata)

    # test with multiple coordinate_labels
    new_coordinate_labels = ["position_x", "position_y"]
    new_locdata = locdata.projection(coordinate_labels=new_coordinate_labels)

    assert all(
        label in new_coordinate_labels for label in new_locdata.coordinate_labels
    )
    assert new_locdata.dimension == len(new_coordinate_labels)
    assert len(new_locdata) == len(locdata)


# locdata with added columns


def test_LocData_add_column_to_dataframe(df_simple):
    # from dataframe
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    dat.dataframe = dat.dataframe.assign(new=np.arange(5))
    assert dat.data.equals(dat.dataframe)

    # from selection
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    sel = LocData.from_selection(locdata=dat, indices=[0, 1, 2])
    sel.dataframe = sel.dataframe.assign(new=np.arange(3))
    assert list(sel.dataframe.columns) == ["new"]
    assert list(sel.references.dataframe.columns) == ["position_x", "position_y"]
    assert all(list(sel.data.columns == ["position_x", "position_y", "new"]))
    sel.reduce()
    assert all(list(sel.dataframe.columns == ["position_x", "position_y", "new"]))

    # sel_2 = LocData.from_selection(locdata=sel, indices=[0, 2])
    # print(sel_2.data)
    # print(sel_2.references.data)
    # print(sel_2.references.references.data)
    # sel_2.reduce()
    # print(sel_2.data)

    # from collection
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    sel_1 = LocData.from_selection(locdata=dat, indices=[0, 1, 2])
    sel_2 = LocData.from_selection(locdata=dat, indices=[3, 4])
    col = LocData.from_collection([sel_1, sel_2], meta=COMMENT_METADATA)
    col.dataframe = col.dataframe.assign(new=np.arange(2))
    # print(col.data.columns)
    assert all(
        list(
            col.data.columns
            == [
                "localization_count",
                "position_x",
                "position_y",
                "region_measure_bb",
                "localization_density_bb",
                "subregion_measure_bb",
                "new",
            ]
        )
    )


# locdata and metadata


def test_LocData_handling_metadata(df_simple):
    dat = LocData.from_dataframe(dataframe=df_simple, meta=COMMENT_METADATA)
    # print(dat.meta)
    # print(dir(metadata_pb2))
    assert dat.meta.identifier
    assert dat.meta.creation_time
    assert dat.meta.source == metadata_pb2.DESIGN
    assert dat.meta.state == metadata_pb2.RAW
    assert dat.meta.history[0].name == "LocData.from_dataframe"
    assert dat.meta.element_count == 5

    dat.meta.comment = "new comment"
    assert dat.meta.comment == "new comment"

    dat.meta.map["variable key"] = "new comment"
    assert dict(dat.meta.map) == {"variable key": "new comment"}

    dat.meta.map["key_2"] = "value_2"
    assert dat.meta.map["key_2"] == "value_2"

    dat = LocData.from_dataframe(
        dataframe=df_simple, meta=dict(comment="yet another comment")
    )
    assert dat.meta.comment == "yet another comment"


# locdata and regions


def test_locdata_region(df_simple):
    region = Rectangle((0, 0), 2, 1, 10)
    dat = LocData.from_dataframe(dataframe=df_simple)
    dat.region = region
    assert isinstance(dat.region, Region)
    assert dat.region.region_measure == 2


# standard LocData fixtures


@pytest.mark.parametrize(
    "fixture_name, expected",
    [
        ("locdata_empty", 0),
        ("locdata_single_localization", 1),
        ("locdata_2d", 6),
        ("locdata_non_standard_index", 6),
    ],
)
def test_standard_locdata_objects(
    locdata_empty,
    locdata_single_localization,
    locdata_2d,
    locdata_non_standard_index,
    fixture_name,
    expected,
):
    dat = eval(fixture_name)
    assert len(dat) == expected


@pytest.mark.parametrize(
    "fixture_name, expected",
    [("locdata_empty", 0), ("locdata_single_localization", 0)],
)
def test_locdata_hulls(
    locdata_empty, locdata_single_localization, fixture_name, expected
):
    dat = eval(fixture_name)
    assert dat.bounding_box.region_measure == 0
    assert dat.oriented_bounding_box.region_measure == 0
    with pytest.warns(UserWarning):
        assert dat.convex_hull is None
    with pytest.raises(AttributeError):
        assert dat.region_measure_ch
    if len(dat) == 0:
        assert isinstance(dat.update_alpha_shape(alpha=1).alpha_shape, AlphaShape)
    else:
        with pytest.warns(UserWarning):
            assert isinstance(dat.update_alpha_shape(alpha=1).alpha_shape, AlphaShape)


@pytest.mark.parametrize(
    "fixture_name, expected",
    [("locdata_empty", 0), ("locdata_single_localization", 0)],
)
def test_locdata_from_selection_exceptions(
    locdata_empty,
    locdata_single_localization,
    locdata_2d,
    locdata_non_standard_index,
    fixture_name,
    expected,
):
    dat = eval(fixture_name)
    new_dat = LocData.from_selection(
        locdata=dat, indices=[1, 3, 4], meta=COMMENT_METADATA
    )
    assert len(new_dat) == expected


@pytest.mark.parametrize(
    "fixture_name, expected", [("locdata_2d", 3), ("locdata_non_standard_index", 3)]
)
def test_locdata_from_selection_(
    locdata_empty,
    locdata_single_localization,
    locdata_2d,
    locdata_non_standard_index,
    fixture_name,
    expected,
):
    dat = eval(fixture_name)
    sel = LocData.from_selection(locdata=dat, indices=[1, 2, 5], meta=COMMENT_METADATA)
    assert len(sel) == expected
    assert set(sel.data.index) == set([1, 2, 5])
    # in coming versions: assert list(sel.data.index) == [2, 1, 5]
    assert sel.references is dat
    assert sel.meta.comment == COMMENT_METADATA.comment


def test_locdata_from_chunks_(locdata_non_standard_index):
    with pytest.raises(ValueError):
        chunk_collection = LocData.from_chunks(locdata=locdata_non_standard_index)

    chunk_collection = LocData.from_chunks(
        locdata=locdata_non_standard_index, chunks=((2, 1), (0, 3))
    )
    assert all(chunk_collection.references[0].data.index == [2, 1])

    chunk_collection = LocData.from_chunks(
        locdata=locdata_non_standard_index, chunk_size=2
    )
    assert all(chunk_collection.references[0].data.index == [2, 1])

    chunk_collection = LocData.from_chunks(
        locdata=locdata_non_standard_index, n_chunks=3
    )
    assert all(chunk_collection.references[0].data.index == [2, 1])

    chunk_collection = LocData.from_chunks(
        locdata=locdata_non_standard_index, n_chunks=3, order="alternating"
    )
    assert all(chunk_collection.references[0].data.index == [2, 10])


@pytest.mark.parametrize(
    "fixture_name, expected",
    [
        ("locdata_empty", (0, (0,), 0, (0,))),
        ("locdata_single_localization", (1, (1,), 1, (1,))),
        ("locdata_2d", (3, (2, 2, 2), 1, (4,))),
        ("locdata_non_standard_index", (3, (2, 2, 2), 1, (4,))),
    ],
)
def test_locdata_from_chunks(
    locdata_empty,
    locdata_single_localization,
    locdata_2d,
    locdata_non_standard_index,
    fixture_name,
    expected,
):
    dat = eval(fixture_name)
    chunk_collection = LocData.from_chunks(
        locdata=dat, chunk_size=2, meta=COMMENT_METADATA
    )
    assert isinstance(chunk_collection.references, list)
    assert len(chunk_collection) == expected[0]
    if len(chunk_collection) != 0:
        assert all(chunk_collection.data.localization_count == expected[1])
    assert chunk_collection.meta.comment == COMMENT_METADATA.comment

    chunk_collection = LocData.from_chunks(
        locdata=dat,
        chunk_size=2,
        order="alternating",
        meta=dict(comment="yet another comment"),
    )
    assert isinstance(chunk_collection.references, list)
    assert len(chunk_collection) == expected[0]
    if len(chunk_collection) != 0:
        assert all(chunk_collection.data.localization_count == expected[1])
    assert chunk_collection.meta.comment == "yet another comment"

    chunk_collection = LocData.from_chunks(locdata=dat, chunk_size=4, drop=True)
    assert isinstance(chunk_collection.references, list)
    assert len(chunk_collection) == expected[2]
    if len(chunk_collection) != 0:
        assert all(chunk_collection.data.localization_count == expected[3])


def test_copy():
    locdata = LocData.from_dataframe(
        dataframe=pd.DataFrame({"col_1": [1, 2, 3], "col_2": ["a", "b", "c"]})
    )
    selection = LocData.from_selection(locdata, indices=[0, 1])
    new_locdata = copy.copy(selection)

    assert new_locdata is not selection
    assert len(selection) == len(new_locdata)
    for attr in ["dataframe", "indices", "references"]:
        assert getattr(selection, attr) is getattr(new_locdata, attr)
    for attr in ["properties", "data", "meta"]:
        assert getattr(selection, attr) is not getattr(new_locdata, attr)
    pd.testing.assert_frame_equal(new_locdata.dataframe, selection.dataframe)
    pd.testing.assert_frame_equal(new_locdata.data, selection.data)
    assert new_locdata.properties == selection.properties
    assert new_locdata.indices == selection.indices
    assert new_locdata.references == selection.references
    assert new_locdata.meta != selection.meta
    assert (
        int(new_locdata.references.meta.identifier)
        == int(selection.meta.identifier) - 1
    )
    assert int(new_locdata.meta.identifier) == int(selection.meta.identifier) + 1
    assert new_locdata.meta.creation_time == selection.meta.creation_time


def test_deepcopy():
    locdata = LocData.from_dataframe(
        dataframe=pd.DataFrame({"col_1": [1, 2, 3], "col_2": ["a", "b", "c"]})
    )
    selection = LocData.from_selection(locdata, indices=[0, 1])
    new_locdata = copy.deepcopy(selection)

    assert new_locdata is not selection
    assert len(selection) == len(new_locdata)
    for attr in ["data", "dataframe", "indices", "meta", "properties", "references"]:
        assert getattr(selection, attr) is not getattr(new_locdata, attr)
    pd.testing.assert_frame_equal(new_locdata.dataframe, selection.dataframe)
    pd.testing.assert_frame_equal(new_locdata.data, selection.data)
    assert new_locdata.properties == selection.properties
    assert new_locdata.indices == selection.indices
    assert new_locdata.references != selection.references
    assert new_locdata.meta != selection.meta
    assert (
        int(new_locdata.references.meta.identifier)
        == int(selection.meta.identifier) + 1
    )
    assert int(new_locdata.meta.identifier) == int(selection.meta.identifier) + 2
    assert new_locdata.meta.creation_time == selection.meta.creation_time


def test_locdata_print_summary(capfd, locdata_2d):
    locdata_2d.print_summary()
    captured = capfd.readouterr()
    for element in ["identifier", "creation_time", "source", "state", "element_count"]:
        assert element in captured.out


def test_locdata_print_meta(capfd, locdata_2d):
    locdata_2d.print_meta()
    captured = capfd.readouterr()
    for element in ["identifier", "creation_time", "source", "state", "element_count"]:
        assert element in captured.out



import tempfile
from pathlib import Path

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
import pytest

import locan
from locan import ROOT_DIR, LocData
from locan.data import metadata_pb2
from locan.data.rois import RoiLegacy_0
from locan.locan_io.locdata.io_locdata import load_txt_file

# fixtures


@pytest.fixture()
def locdata():
    locdata_dict = {
        "position_x": [0, 1, 2, 3, 0, 1, 4, 5],
        "position_y": [0, 1, 2, 3, 1, 4, 5, 1],
        "position_z": [0, 1, 2, 3, 4, 4, 4, 5],
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def points():
    return np.array([[0, 0], [0.5, 0.5], [100, 100], [1, 1], [1.1, 0.9]])


# tests


def test_Roi_2d(locdata_2d):
    roi = RoiLegacy_0(
        reference=locdata_2d, region_specs=((0, 0), 2, 2, 0), region_type="rectangle"
    )
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is roi.region
    assert isinstance(roi.region, locan.RoiRegion)
    roi.region = locan.RoiRegion(
        region_specs=((0, 0), 2, 2, 0), region_type="rectangle"
    )
    assert isinstance(roi.region, locan.RoiRegion)

    roi = RoiLegacy_0(
        reference=locdata_2d,
        region_specs=((0, 0), 4, 4, 0),
        region_type="rectangle",
        properties_for_roi=("position_y", "frame"),
    )
    new_dat = roi.locdata()
    assert len(new_dat) == 2
    assert new_dat.region is None

    locdata_empty = LocData()
    roi_empty = RoiLegacy_0(
        reference=locdata_empty,
        region_specs=((0, 0), 2, 1, 10),
        region_type="rectangle",
    )
    empty_dat = roi_empty.locdata()
    assert len(empty_dat) == 0
    assert new_dat.region is None


def test_Roi_3d(locdata_3d):
    roi = RoiLegacy_0(
        reference=locdata_3d, region_specs=((0, 0), 2, 2, 0), region_type="rectangle"
    )
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is None
    assert isinstance(roi.region, locan.RoiRegion)
    roi.region = locan.RoiRegion(
        region_specs=((0, 0), 2, 1, 0), region_type="rectangle"
    )
    assert isinstance(roi.region, locan.RoiRegion)

    roi = RoiLegacy_0(
        reference=locdata_3d,
        region_specs=((0, 0), 2, 3, 0),
        region_type="rectangle",
        properties_for_roi=("position_y", "position_z"),
    )
    new_dat = roi.locdata()
    assert len(new_dat) == 1
    assert new_dat.region is None


def test_pickling_locdata_from_Roi(locdata_2d):
    import pickle

    roi = RoiLegacy_0(
        reference=locdata_2d, region_specs=((0, 0), 2, 1, 0), region_type="rectangle"
    )
    new_dat = roi.locdata()

    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / "pickled_locdata.pickle"
        with open(file_path, "wb") as file:
            pickle.dump(new_dat, file, pickle.HIGHEST_PROTOCOL)
        with open(file_path, "rb") as file:
            locdata = pickle.load(file)
        assert len(new_dat) == len(locdata)
        assert isinstance(locdata.meta, metadata_pb2.Metadata)


def test_Roi_io(locdata):
    with tempfile.TemporaryDirectory() as tmp_directory:
        file_path = Path(tmp_directory) / "roi.yaml"
        # file_path = ROOT_DIR / 'tests/test_data/roi.yaml'

        roi = RoiLegacy_0(
            reference=locdata, region_type="rectangle", region_specs=((0, 0), 2, 1, 10)
        )
        with pytest.deprecated_call():
            roi.to_yaml(path=file_path)

        roi_new = RoiLegacy_0.from_yaml(path=file_path)
        assert roi_new.reference is None

        roi = RoiLegacy_0(
            reference=dict(
                file_path=ROOT_DIR / "tests/test_data/five_blobs.txt", file_type=1
            ),
            region_type="rectangle",
            region_specs=((0, 0), 2, 1, 10),
        )
        assert isinstance(roi.reference, (metadata_pb2.Metadata, Path))
        with pytest.deprecated_call():
            roi.to_yaml(path=file_path)

        roi_new = RoiLegacy_0.from_yaml(path=file_path)
        assert roi_new

        # test region specs with numpy floats
        roi = RoiLegacy_0(
            reference=dict(file_path=file_path, file_type=1),
            region_type="rectangle",
            region_specs=(np.array([0, 0], dtype=float), float(2), float(1), float(10)),
        )
        assert isinstance(roi.reference, metadata_pb2.Metadata)
        with pytest.deprecated_call():
            roi.to_yaml(path=file_path)

        roi_new = RoiLegacy_0.from_yaml(path=file_path)
        assert roi_new

        meta_ = metadata_pb2.Metadata()
        meta_.file.path = str(file_path)
        meta_.file.type = 1
        locdata_2 = LocData.from_selection(locdata, meta=meta_)
        roi = RoiLegacy_0(
            reference=locdata_2,
            region_type="rectangle",
            region_specs=((0, 0), 2, 1, 10),
        )
        assert isinstance(roi.reference.meta, metadata_pb2.Metadata)
        with pytest.deprecated_call():
            roi.to_yaml(path=file_path)

        roi_new = RoiLegacy_0.from_yaml(path=file_path)
        assert roi_new
        assert isinstance(roi_new.reference, metadata_pb2.Metadata)


def test_roi_locdata_from_file():
    locdata = load_txt_file(path=ROOT_DIR / "tests/test_data/five_blobs_3D.txt")

    roi = RoiLegacy_0(
        reference=locdata,
        region_type="rectangle",
        region_specs=((100, 500), 200, 200, 10),
    )
    dat = roi.locdata()
    assert len(dat) == 10

    roi = RoiLegacy_0(
        reference=locdata,
        region_type="rectangle",
        region_specs=((100, 100), 200, 200, 10),
        properties_for_roi=["position_x", "position_z"],
    )
    dat = roi.locdata()
    assert len(dat) == 10


def test_as_artist():
    roi_rectangle = RoiLegacy_0(
        reference=locdata, region_specs=((0, 0), 0.5, 0.5, 10), region_type="rectangle"
    )
    roi_ellipse = RoiLegacy_0(
        reference=locdata,
        region_specs=((0.5, 0.5), 0.2, 0.3, 10),
        region_type="ellipse",
    )
    roi_polygon = RoiLegacy_0(
        reference=locdata,
        region_specs=((0, 0.8), (0.1, 0.9), (0.6, 0.9), (0.7, 0.7), (0, 0.8)),
        region_type="polygon",
    )
    fig, ax = plt.subplots()
    ax.add_patch(roi_rectangle._region.as_artist())
    ax.add_patch(roi_ellipse._region.as_artist())
    ax.add_patch(roi_polygon._region.as_artist())
    # plt.show()

    plt.close("all")



"""

Compute localizations per frame.

"""
import logging
from dataclasses import dataclass

import matplotlib.pyplot as plt
import numpy as np
import pandas as pd
from scipy import stats

from locan.analysis.analysis_base import _Analysis

__all__ = ["LocalizationsPerFrame"]

logger = logging.getLogger(__name__)


# The algorithms


def _localizations_per_frame(
    locdata, norm=None, time_delta="integration_time", resample=None, **kwargs
):
    """
    Compute localizations per frame.

    Parameters
    ----------
    locdata : LocData or array-like
       Points in time: either localization data that contains a column `frame`
       or an array with time points.
    norm : int, float, str, None
        Normalization factor that can be None, a number, or another property in
         `locdata`.
    time_delta : int, float, str, pd.Timedelta, None
        Time per frame in milliseconds. String must specify the unit like
        "10ms".
         For "integration_time" the time is taken from
        locdata.meta.experiment.setups[0].optical_units[0].detection.camera.integration_time
    resample : DateOffset, Timedelta or str
        Parameter for :func:`pandas.Series.resample`: The offset string or
        object representing target conversion.
    kwargs : dict
        Other parameters passed to :func:`pandas.Series.resample`.


    Returns:
    --------
    pandas.Series
    """
    # normalization
    if norm is None:
        normalization_factor = 1
        series_name = "n_localizations"
    elif isinstance(norm, str):
        normalization_factor = locdata.properties[norm]
        series_name = "n_localizations / " + norm
    elif isinstance(norm, (int, float)):
        normalization_factor = norm
        series_name = f"n_localizations / {norm}"
    else:
        raise TypeError(
            "normalization should be None, a number or a valid property name."
        )

    try:
        frames_ = locdata.data.frame.astype(int)
    except AttributeError:
        frames_ = np.asarray(locdata)

    frames, frame_counts = np.unique(frames_, return_counts=True)
    series = pd.Series(frame_counts, index=frames, dtype=float) / normalization_factor
    series.index.name = "frame"
    series.name = series_name

    if time_delta is None:
        pass
    elif isinstance(time_delta, pd.Timedelta):
        series.index = frames * time_delta
        series.index.name = "time"
    elif isinstance(time_delta, (int, float)):
        series.index = pd.to_timedelta(frames * time_delta, unit="ms")
        series.index.name = "time"
    elif time_delta == "integration_time":
        try:
            time_delta = (
                locdata.meta.experiment.setups[0]
                .optical_units[0]
                .detection.camera.integration_time.ToTimedelta()
            )
            series.index = frames * time_delta
            series.index.name = "time"
        except (IndexError, AttributeError):
            logger.warning(
                "integration_time not available in locdata.meta - frames used instead."
            )
    elif isinstance(time_delta, str):
        time_delta = pd.Timedelta(time_delta)
        series.index = frames * time_delta
        series.index.name = "time"
    else:
        raise ValueError("The input for time_delta is not implemented.")

    if resample is not None:
        series = series.resample(resample, **kwargs).sum()

    return series


# The specific analysis classes


@dataclass(repr=False)
class _Results:
    time_series: pd.Series

    def accumulation_time(self, fraction=0.5) -> int:
        normalized_cumulative_time_trace = (
            self.time_series.cumsum() / self.time_series.sum()
        )
        accumulation_time = normalized_cumulative_time_trace.gt(fraction).idxmax()
        return accumulation_time


class LocalizationsPerFrame(_Analysis):
    """
    Compute localizations per frame.

    Parameters
    ----------
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    norm : int, float, str, None
        Normalization factor that can be None, a number, or another property
        in `locdata`.
    time_delta : int, float, str, None
        Time per frame or "integration_time" for which the time is taken from
        locdata.meta.experiment.setups[0].optical_units[0].detection.camera.integration_time
    resample : DateOffset, Timedelta or str
        Parameter for :func:`pandas.Series.resample`: The offset string or
        object representing target conversion.
    kwargs : dict
        Other parameters passed to :func:`pandas.Series.resample`.

    Attributes
    ----------
    count : int
        A counter for counting instantiations.
    parameter : dict
        A dictionary with all settings for the current computation.
    meta : locan.analysis.metadata_analysis_pb2.AMetadata
        Metadata about the current analysis routine.
    results : pandas.Series
        Computed results.
    distribution_statistics : Distribution_fits object, None
        Distribution parameters derived from MLE fitting of results.
    """

    count = 0

    def __init__(
        self,
        meta=None,
        norm=None,
        time_delta="integration_time",
        resample=None,
        **kwargs,
    ):
        super().__init__(
            meta=meta, norm=norm, time_delta=time_delta, resample=resample, **kwargs
        )
        self.results = None
        self.distribution_statistics = None

    def compute(self, locdata):
        """
        Run the computation.

        Parameters
        ----------
        locdata : LocData or array-like
           Points in time: either localization data that contains a column
           `frame` or an array with time points.

        Returns
        -------
        Analysis class
           Returns the Analysis class object (self).
        """
        if not len(locdata):
            logger.warning("Locdata is empty.")
            self.distribution_statistics = None
            self.results = None
            return self

        self.distribution_statistics = None
        self.results = _Results(
            time_series=_localizations_per_frame(locdata=locdata, **self.parameter)
        )
        return self

    def fit_distributions(self, **kwargs):
        """
        Fit probability density functions to the distributions of `
        loc_property` values in the results using MLE (scipy.stats).

        Parameters
        ----------
        loc_property : str
            The LocData property for which to fit an appropriate distribution;
            if None all plots are shown.
        """
        if self:
            self.distribution_statistics = _DistributionFits(self)
            self.distribution_statistics.fit(**kwargs)
        else:
            logger.warning("No results available to fit.")

    def plot(self, ax=None, window=1, cumulative=False, normalize=False, **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        running average of results over window size.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        window: int
            Window for running average that is applied before plotting.
        cumulative : bool
            Plot the cumulated results if true.
        normalize : bool
            Normalize cumulative plot to the last value
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        series = self.results.time_series

        # prepare plot
        if cumulative and normalize:
            _results = series.cumsum() / series.sum()
        elif cumulative and not normalize:
            _results = series.cumsum()
        elif not cumulative:
            _results = series
        else:
            return ax

        _results.rolling(window=window, center=True).mean().plot(ax=ax, **kwargs)

        ax.set(
            title=f"Localizations per Frame\n (window={window})",
            xlabel=series.index.name,
            ylabel=f"{series.name} (cumulative)" if cumulative else series.name,
        )

        return ax

    def hist(self, ax=None, fit=True, bins="auto", **kwargs):
        """
        Provide histogram as :class:`matplotlib.axes.Axes` object showing
        hist(results).

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image
        bins : float
            Bin specifications (passed to matplotlib.hist).
        fit: Bool
            Flag indicating if distributions fit are shown.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.hist`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if not self:
            return ax

        series = self.results.time_series

        ax.hist(
            series.values, bins=bins, **dict(dict(density=True, log=False), **kwargs)
        )
        ax.set(title="Localizations per Frame", xlabel=series.name, ylabel="PDF")

        if fit:
            if isinstance(self.distribution_statistics, _DistributionFits):
                self.distribution_statistics.plot(ax=ax)
            else:
                self.fit_distributions()
                self.distribution_statistics.plot(ax=ax)

        return ax


# todo: add fit function


class _DistributionFits:
    """
    Handle for distribution fits.

    It holds the statistical parameters derived by fitting the result
    distributions using MLE (scipy.stats).
    Statistical parameters are defined as described in
    :ref:(https://docs.scipy.org/doc/scipy/reference/tutorial/stats/continuous.html)

    Parameters
    ----------
    analyis_class : LocalizationPrecision object
        The analysis class with result data to fit.

    Attributes
    ----------
    analyis_class : LocalizationPrecision
        The analysis class with result data to fit.
    loc_property : str
        The LocData property for which to fit an appropriate distribution
    distribution : str, scipy.stats.distribution
        Distribution model to fit.
    parameters :
    """

    def __init__(self, analysis_class):
        self.analysis_class = analysis_class
        self.loc_property = self.analysis_class.results.time_series.name
        self.distribution = None
        self.parameters = []

    def fit(self, distribution=stats.norm, **kwargs):
        """
        Fit scipy.stats.distribution to analysis_class.results[loc_property].

        If with_constraints is true we put the following constraints on the fit
        procedure:
        If distribution is expon then
        floc=np.min(self.analysis_class.results[self.loc_property].values).

        Parameters
        ----------
        distribution : str, scipy.stats.distribution
            Distribution model to fit.
        kwargs : dict
            Other parameters are passed to
            :func:`scipy.stat.distribution.fit()`.
        """
        if self.analysis_class.results is None:
            return

        self.distribution = distribution

        loc, scale = self.distribution.fit(
            self.analysis_class.results.time_series.values, **kwargs
        )
        self.parameters.extend(
            [self.loc_property + "_center", self.loc_property + "_sigma"]
        )
        setattr(self, self.loc_property + "_center", loc)
        setattr(self, self.loc_property + "_sigma", scale)

    def plot(self, ax=None, **kwargs):
        """
        Provide plot as :class:`matplotlib.axes.Axes` object showing the
        probability distribution functions of fitted results.

        Parameters
        ----------
        ax : :class:`matplotlib.axes.Axes`
            The axes on which to show the image.
        kwargs : dict
            Other parameters passed to :func:`matplotlib.pyplot.plot`.

        Returns
        -------
        :class:`matplotlib.axes.Axes`
            Axes object with the plot.
        """
        if ax is None:
            ax = plt.gca()

        if self.distribution is None:
            return ax

        # plot fit curve
        _center, _sigma = self.parameter_dict().values()

        x_values = np.linspace(
            self.distribution.ppf(0.001, loc=_center, scale=_sigma),
            self.distribution.ppf(0.999, loc=_center, scale=_sigma),
            100,
        )
        ax.plot(
            x_values,
            self.distribution.pdf(x_values, loc=_center, scale=_sigma),
            "r-",
            **dict(
                dict(lw=3, alpha=0.6, label=str(self.distribution.name) + " pdf"),
                **kwargs,
            ),
        )
        return ax

    def parameter_dict(self):
        """Dictionary of fitted parameters."""
        if self.parameters is None:
            return None
        else:
            return {k: self.__dict__[k] for k in self.parameters}



from copy import deepcopy

import numpy as np
import pytest

from locan import (
    Bins,
    cluster_by_bin,
    cluster_dbscan,
    cluster_hdbscan,
    serial_clustering,
)
from locan.dependencies import HAS_DEPENDENCY


# tests hdbscan
@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test_cluster_hdbscan_2d(locdata_two_cluster_2d):
    noise, clust = cluster_hdbscan(
        locdata_two_cluster_2d, min_cluster_size=2, allow_single_cluster=False
    )
    assert noise is None
    assert len(clust) == 2
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_hdbscan(
        locdata_two_cluster_2d, min_cluster_size=5, allow_single_cluster=True
    )
    assert len(clust) == 1
    assert clust.data.localization_count[0] == 6
    assert all(np.in1d(clust.references[0].data.cluster_label, [1, 2]))
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_hdbscan(
        locdata_two_cluster_2d,
        loc_properties=["position_x"],
        min_cluster_size=2,
        allow_single_cluster=False,
    )
    assert noise is None
    assert len(clust) == 2
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_2d.region


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test_cluster_hdbscan_2d_with_noise(locdata_two_cluster_with_noise_2d):
    noise, clust = cluster_hdbscan(
        locdata_two_cluster_with_noise_2d,
        min_cluster_size=2,
        allow_single_cluster=False,
    )
    assert len(noise) == 1
    assert len(clust) == 2
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region

    noise, clust = cluster_hdbscan(
        locdata_two_cluster_with_noise_2d, min_cluster_size=5, allow_single_cluster=True
    )
    assert len(noise) == 1
    assert len(clust) == 1
    assert clust.data.localization_count[0] == 6
    assert all(np.in1d(clust.references[0].data.cluster_label, [1, 2]))
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
def test_cluster_hdbscan_with_shuffled_index(locdata_two_cluster_with_noise_2d):
    locdata = deepcopy(locdata_two_cluster_with_noise_2d)
    new_index = list(locdata.data.index)
    np.random.shuffle(new_index)
    locdata.data.index = new_index
    noise, clust = cluster_hdbscan(
        locdata, min_cluster_size=2, allow_single_cluster=False
    )
    assert len(noise) == 1
    assert len(clust) == 2
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)


@pytest.mark.skipif(not HAS_DEPENDENCY["hdbscan"], reason="Test requires hdbscan.")
@pytest.mark.parametrize(
    "fixture_name, expected",
    [("locdata_empty", (0, 0)), ("locdata_single_localization", (1, 0))],
)
def test_cluster_hdbscan_empty_locdata(
    locdata_empty, locdata_single_localization, fixture_name, expected
):
    locdata = eval(fixture_name)
    noise, clust = cluster_hdbscan(
        locdata, min_cluster_size=2, allow_single_cluster=False
    )
    assert len(noise) == expected[0]
    assert len(clust) == expected[1]


# tests dbscan


def test_cluster_dbscan_2d(locdata_two_cluster_2d):
    noise, clust = cluster_dbscan(locdata_two_cluster_2d, eps=2, min_samples=1)
    assert noise is None
    assert len(clust) == 2
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_dbscan(locdata_two_cluster_2d, eps=20, min_samples=1)
    assert len(clust) == 1
    assert clust.data.localization_count[0] == 6
    assert all(np.in1d(clust.references[0].data.cluster_label, [1, 2]))
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_dbscan(
        locdata_two_cluster_2d, loc_properties=["position_x"], eps=2, min_samples=1
    )
    assert noise is None
    assert len(clust) == 2
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_2d.region

    noise, clust = cluster_dbscan(locdata_two_cluster_2d, eps=0.2, min_samples=10)
    assert len(noise) == len(locdata_two_cluster_2d)
    assert len(clust) == 0
    assert noise.region == locdata_two_cluster_2d.region
    assert clust.region is None


def test_cluster_dbscan_2d_with_noise(locdata_two_cluster_with_noise_2d):
    noise, clust = cluster_dbscan(
        locdata_two_cluster_with_noise_2d, eps=2, min_samples=1
    )
    assert noise is None
    assert len(clust) == 3
    assert all(clust.data.localization_count == [3, 3, 1])
    assert all(clust.references[0].data.cluster_label == 1)
    assert clust.region == locdata_two_cluster_with_noise_2d.region

    noise, clust = cluster_dbscan(
        locdata_two_cluster_with_noise_2d, eps=2, min_samples=2
    )
    assert len(noise) == 1
    assert len(clust) == 2
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region

    noise, clust = cluster_dbscan(
        locdata_two_cluster_with_noise_2d, eps=20, min_samples=2
    )
    assert len(noise) == 1
    assert len(clust) == 1
    assert clust.data.localization_count[0] == 6
    assert all(np.in1d(clust.references[0].data.cluster_label, [1, 2]))
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region


def test_cluster_dbscan_with_shuffled_index(locdata_two_cluster_with_noise_2d):
    locdata = deepcopy(locdata_two_cluster_with_noise_2d)
    new_index = list(locdata.data.index)
    np.random.shuffle(new_index)
    locdata.data.index = new_index
    noise, clust = cluster_dbscan(locdata, eps=2, min_samples=2)
    assert len(noise) == 1
    assert len(clust) == 2
    assert all(clust.data.localization_count == [3, 3])
    assert all(clust.references[0].data.cluster_label == 1)  #
    assert noise.region == locdata_two_cluster_with_noise_2d.region
    assert clust.region == locdata_two_cluster_with_noise_2d.region


@pytest.mark.parametrize(
    "fixture_name, expected",
    [("locdata_empty", (0, 0)), ("locdata_single_localization", (1, 0))],
)
def test_cluster_dbscan_empty_locdata(
    locdata_empty, locdata_single_localization, fixture_name, expected
):
    locdata = eval(fixture_name)
    noise, clust = cluster_dbscan(locdata, eps=2, min_samples=2)
    assert len(noise) == expected[0]
    assert len(clust) == expected[1]


# tests serial_clustering


def test_serial_clustering(locdata_two_cluster_with_noise_2d):
    noise, clust = serial_clustering(
        locdata_two_cluster_with_noise_2d,
        cluster_dbscan,
        parameter_lists=dict(eps=[2, 20], min_samples=[2, 5]),
    )
    assert len(noise) == 4
    assert len(clust) == 4


# tests cluster_by_bin


@pytest.mark.parametrize(
    "fixture_name, expected",
    [
        ("locdata_empty", (np.array([]), 0, np.array([]))),
        ("locdata_single_localization", (np.array([[1, 1]]), 1, np.array([1]))),
    ],
)
def test_cluster_by_bin_empty_locdata(
    locdata_empty, locdata_single_localization, fixture_name, expected
):
    locdata = eval(fixture_name)
    bins, bin_indices, collection, counts = cluster_by_bin(
        locdata,
        loc_properties=["position_x", "position_y"],
        bin_size=2,
        return_counts=True,
    )
    assert bins is None or isinstance(bins, Bins)
    assert np.array_equal(bin_indices, expected[0])
    assert len(bin_indices) == len(collection)
    assert len(collection) == expected[1]
    assert np.array_equal(counts, expected[2])


def test_cluster_by_bin(locdata_2d):
    bins, bin_indices, collection, counts = cluster_by_bin(
        locdata_2d, loc_properties=["position_x", "position_y"], bin_size=5
    )
    assert bins.bin_size == (4, 5)
    assert np.array_equal(bin_indices, [[1, 1], [1, 2], [2, 1]])
    assert len(bin_indices) == len(collection)
    assert counts is None

    # test with min_samples
    bins, bin_indices, collection, counts = cluster_by_bin(
        locdata_2d,
        loc_properties=["position_x", "position_y"],
        bin_size=5,
        min_samples=2,
        return_counts=True,
    )
    assert bins.bin_size == (4, 5)
    assert np.array_equal(bin_indices, [[1, 1]])
    assert len(bin_indices) == len(collection)
    assert np.array_equal(counts, [4])



"""

Utility functions for file input/output.

"""
import logging
from pathlib import Path

__all__ = ["find_pattern_upstream"]

logger = logging.getLogger(__name__)


def find_pattern_upstream(sub_directory, pattern, directory=None):
    """
    Search for first upstream parent of sub_directory that contains pattern.
    Return first pattern found.
    Return None if no pattern has been found when parent equals directory.

    Parameters
    ----------
    sub_directory : str | bytes | os.PathLike
        Directory or file path to start with.
    pattern : str
        The same pattern as in :func:`Path.glob`.
    directory : str | bytes | os.PathLike
        Directory in which to stop the search.

    Returns
    -------
    Path | None
    """
    sub_directory = Path(sub_directory).resolve(strict=True)

    if directory is None:
        directory = sub_directory.anchor

    for parent in sub_directory.parents:
        file_list = list(parent.glob(pattern))
        if file_list:
            return file_list[0]
        if parent == directory:
            return None



import numpy as np
import pytest

from locan.data.register import register_cc, register_icp
from locan.data.transform import transform_affine
from locan.dependencies import HAS_DEPENDENCY


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_register_icp_2d(locdata_blobs_2d):
    locdata_2d_transformed = transform_affine(locdata_blobs_2d, offset=(100, 100))
    offset_target = np.array([100.0, 100.0])
    matrix_target = np.array([[1, 0], [0, 1]])

    matrix, offset = register_icp(
        locdata_blobs_2d, locdata_2d_transformed, verbose=False
    )
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = register_icp(
        locdata_blobs_2d,
        locdata_2d_transformed,
        pre_translation=(-90, -90),
        verbose=False,
    )
    offset_target = np.array([10.0, 10.0])
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)


@pytest.mark.skipif(not HAS_DEPENDENCY["open3d"], reason="Test requires open3d.")
def test_register_icp_3d(locdata_blobs_3d):
    locdata_3d_transformed = transform_affine(locdata_blobs_3d, offset=(100, 100, 100))
    offset_target = np.array([100.0, 100.0, 100.0])
    matrix_target = np.array([[1, 0, 0], [0, 1, 0], [0, 0, 1]])

    matrix, offset = register_icp(
        locdata_blobs_3d, locdata_3d_transformed, verbose=False
    )
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = register_icp(
        locdata_blobs_3d,
        locdata_3d_transformed,
        pre_translation=(-90, -90, -90),
        verbose=False,
    )
    offset_target = np.array([10.0, 10.0, 10.0])
    assert np.allclose(np.array(offset), offset_target)
    assert np.allclose(matrix, matrix_target)


def test_register_cc(locdata_blobs_2d):
    locdata_2d_transformed = transform_affine(locdata_blobs_2d, offset=(100, 50))
    offset_target = np.array([100.0, 50.0])
    matrix_target = np.array([[1, 0], [0, 1]])

    matrix, offset = register_cc(
        locdata_blobs_2d, locdata_2d_transformed, bin_size=50, verbose=False
    )
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)

    matrix, offset = register_cc(
        locdata_blobs_2d, locdata_2d_transformed, bin_size=(10, 50), verbose=False
    )
    assert np.allclose(np.array(offset), offset_target, atol=5)
    assert np.allclose(matrix, matrix_target)



import numpy as np
import pandas as pd
import pytest
from scipy import stats

from locan import LocData
from locan.analysis import BlinkStatistics
from locan.analysis.blinking import _blink_statistics, _DistributionFits


def test__blink_statistics_0():
    # frame with on and off periods up to three frames and starting with one-frame on-period.
    frames = np.array([0, 4, 6, 7, 8, 12, 13])

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [1, 1, 3, 2])
    assert np.array_equal(results["off_periods"], [3, 1, 3])
    assert np.array_equal(results["on_periods_frame"], [0, 4, 6, 12])
    assert np.array_equal(results["off_periods_frame"], [1, 5, 9])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0], [1], [2, 3, 4], [5, 6]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=1, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [1, 5, 2])
    assert np.array_equal(results["off_periods"], [3, 3])
    assert np.array_equal(results["on_periods_frame"], [0, 4, 12])
    assert np.array_equal(results["off_periods_frame"], [1, 9])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0], [1, 2, 3, 4], [5, 6]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [14])
    assert np.array_equal(results["off_periods"], [])
    assert np.array_equal(results["on_periods_frame"], [0])
    assert np.array_equal(results["off_periods_frame"], [])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(results["on_periods_indices"], [[0, 1, 2, 3, 4, 5, 6]])
        ]
    )


def test__blink_statistics_1():
    # frame with on and off periods up to three frames and starting with two-frame on-period.
    frames = np.array([0, 1, 3, 6, 7, 8, 12, 13])

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [2, 1, 3, 2])
    assert np.array_equal(results["off_periods"], [1, 2, 3])
    assert np.array_equal(results["on_periods_frame"], [0, 3, 6, 12])
    assert np.array_equal(results["off_periods_frame"], [2, 4, 9])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1], [2], [3, 4, 5], [6, 7]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=1, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [4, 3, 2])
    assert np.array_equal(results["off_periods"], [2, 3])
    assert np.array_equal(results["on_periods_frame"], [0, 6, 12])
    assert np.array_equal(results["off_periods_frame"], [4, 9])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1, 2], [3, 4, 5], [6, 7]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [14])
    assert np.array_equal(results["off_periods"], [])
    assert np.array_equal(results["on_periods_frame"], [0])
    assert np.array_equal(results["off_periods_frame"], [])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1, 2, 3, 4, 5, 6, 7]]
            )
        ]
    )


def test__blink_statistics_2():
    # frame with on and off periods up to three frames and starting with two-frame on-period.
    frames = np.array([0, 1, 3, 6, 7, 8, 12, 13]) + 1

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [2, 1, 3, 2])
    assert np.array_equal(results["off_periods"], [1, 1, 2, 3])
    assert np.array_equal(results["on_periods_frame"], [1, 4, 7, 13])
    assert np.array_equal(results["off_periods_frame"], [0, 3, 5, 10])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1], [2], [3, 4, 5], [6, 7]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=1, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [5, 3, 2])
    assert np.array_equal(results["off_periods"], [2, 3])
    assert np.array_equal(results["on_periods_frame"], [0, 7, 13])
    assert np.array_equal(results["off_periods_frame"], [5, 10])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1, 2], [3, 4, 5], [6, 7]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [15])
    assert np.array_equal(results["off_periods"], [])
    assert np.array_equal(results["on_periods_frame"], [0])
    assert np.array_equal(results["off_periods_frame"], [])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1, 2, 3, 4, 5, 6, 7]]
            )
        ]
    )


def test__blink_statistics_3():
    # frame with on and off periods up to three frames and starting with off-period.
    frames = np.array([0, 1, 4, 6, 7, 8, 12, 13]) + 4

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [2, 1, 3, 2])
    assert np.array_equal(results["off_periods"], [4, 2, 1, 3])
    assert np.array_equal(results["on_periods_frame"], [4, 8, 10, 16])
    assert np.array_equal(results["off_periods_frame"], [0, 6, 9, 13])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1], [2], [3, 4, 5], [6, 7]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=True)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [2, 1, 3, 2])
    assert np.array_equal(results["off_periods"], [2, 1, 3])
    assert np.array_equal(results["on_periods_frame"], [4, 8, 10, 16])
    assert np.array_equal(results["off_periods_frame"], [6, 9, 13])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1], [2], [3, 4, 5], [6, 7]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=2, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [9, 2])
    assert np.array_equal(results["off_periods"], [4, 3])
    assert np.array_equal(results["on_periods_frame"], [4, 16])
    assert np.array_equal(results["off_periods_frame"], [0, 13])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1, 2, 3, 4, 5], [6, 7]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=2, remove_heading_off_periods=True)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [9, 2])
    assert np.array_equal(results["off_periods"], [3])
    assert np.array_equal(results["on_periods_frame"], [4, 16])
    assert np.array_equal(results["off_periods_frame"], [13])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1, 2, 3, 4, 5], [6, 7]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [18])
    assert np.array_equal(results["off_periods"], [])
    assert np.array_equal(results["on_periods_frame"], [0])
    assert np.array_equal(results["off_periods_frame"], [])
    assert np.array_equal(results["on_periods_indices"], [[0, 1, 2, 3, 4, 5, 6, 7]])


def test__blink_statistics_4():
    # frame with on and off periods up to three frames and starting with off-period.
    frames = np.array([0, 1, 4, 6, 12, 13]) + 2

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [2, 1, 1, 2])
    assert np.array_equal(results["off_periods"], [2, 2, 1, 5])
    assert np.array_equal(results["on_periods_frame"], [2, 6, 8, 14])
    assert np.array_equal(results["off_periods_frame"], [0, 4, 7, 9])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1], [2], [3], [4, 5]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=True)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [2, 1, 1, 2])
    assert np.array_equal(results["off_periods"], [2, 1, 5])
    assert np.array_equal(results["on_periods_frame"], [2, 6, 8, 14])
    assert np.array_equal(results["off_periods_frame"], [4, 7, 9])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1], [2], [3], [4, 5]]
            )
        ]
    )

    results = _blink_statistics(frames, memory=3, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [9, 2])
    assert np.array_equal(results["off_periods"], [5])
    assert np.array_equal(results["on_periods_frame"], [0, 14])
    assert np.array_equal(results["off_periods_frame"], [9])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(results["on_periods_indices"], [[0, 1, 2, 3], [4, 5]])
        ]
    )

    results = _blink_statistics(frames, memory=3, remove_heading_off_periods=True)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [7, 2])
    assert np.array_equal(results["off_periods"], [5])
    assert np.array_equal(results["on_periods_frame"], [2, 14])
    assert np.array_equal(results["off_periods_frame"], [9])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(results["on_periods_indices"], [[0, 1, 2, 3], [4, 5]])
        ]
    )

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [16])
    assert np.array_equal(results["off_periods"], [])
    assert np.array_equal(results["on_periods_frame"], [0])
    assert np.array_equal(results["off_periods_frame"], [])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(results["on_periods_indices"], [[0, 1, 2, 3, 4, 5]])
        ]
    )


def test__blink_statistics_5(caplog):
    # frame with on and off periods including repeated frames.
    frames = np.array([0, 1, 4, 4, 6, 7, 8, 12, 12, 13]) + 4

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [2, 1, 3, 2])
    assert np.array_equal(results["off_periods"], [4, 2, 1, 3])
    assert np.array_equal(results["on_periods_frame"], [4, 8, 10, 16])
    assert np.array_equal(results["off_periods_frame"], [0, 6, 9, 13])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1], [2], [3, 4, 5], [6, 7]]
            )
        ]
    )
    assert caplog.record_tuples == [
        (
            "locan.analysis.blinking",
            30,
            "There are 2 duplicated frames found that will be ignored.",
        )
    ]

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [18])
    assert np.array_equal(results["off_periods"], [])
    assert np.array_equal(results["on_periods_frame"], [0])
    assert np.array_equal(results["off_periods_frame"], [])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(
                results["on_periods_indices"], [[0, 1, 2, 3, 4, 5, 6, 7]]
            )
        ]
    )


def test__blink_statistics_6():
    # frame with on and off periods up to three frames and starting with one-frame on-period.
    frames = np.array([0, 2, 3, 9])

    results = _blink_statistics(frames, memory=0, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [1, 2, 1])
    assert np.array_equal(results["off_periods"], [1, 5])
    assert np.array_equal(results["on_periods_frame"], [0, 2, 9])
    assert np.array_equal(results["off_periods_frame"], [1, 4])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(results["on_periods_indices"], [[0], [1, 2], [3]])
        ]
    )

    results = _blink_statistics(frames, memory=10, remove_heading_off_periods=False)
    assert len(results["on_periods"]) == len(results["on_periods_frame"])
    assert len(results["off_periods"]) == len(results["off_periods_frame"])
    assert np.array_equal(results["on_periods"], [10])
    assert np.array_equal(results["off_periods"], [])
    assert np.array_equal(results["on_periods_frame"], [0])
    assert np.array_equal(results["off_periods_frame"], [])
    assert all(
        [
            np.array_equal(one, two)
            for one, two in zip(results["on_periods_indices"], [[0, 1, 2, 3]])
        ]
    )


@pytest.fixture()
def locdata_simple():
    locdata_dict = {
        "position_x": [0, 0, 1, 4, 5],
        "position_y": [0, 1, 3, 4, 1],
        "intensity": [0, 1, 3, 4, 1],
        "psf_sigma_x": [100, 100, 100, 100, 100],
    }
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def locdata_with_zero_frame():
    locdata_dict = {"frame": [0, 1, 2, 4, 10, 11, 14]}
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def locdata_without_zero_frame():
    locdata_dict = {"frame": [1, 2, 4, 10, 11, 14]}
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


@pytest.fixture()
def locdata_with_repetitions():
    locdata_dict = {"frame": [2, 2, 2, 4, 4, 14]}
    return LocData(dataframe=pd.DataFrame.from_dict(locdata_dict))


def test_blink_statistics(locdata_with_zero_frame, locdata_without_zero_frame):
    bs = _blink_statistics(
        locdata_with_zero_frame, memory=0, remove_heading_off_periods=False
    )
    assert all(bs["on_periods"] == [3, 1, 2, 1])
    assert all(bs["off_periods"] == [1, 5, 2])

    bs = _blink_statistics(
        locdata_with_zero_frame.data.frame.values,
        memory=0,
        remove_heading_off_periods=False,
    )
    assert all(bs["on_periods"] == [3, 1, 2, 1])
    assert all(bs["off_periods"] == [1, 5, 2])

    bs = _blink_statistics(
        locdata_without_zero_frame, memory=0, remove_heading_off_periods=False
    )
    assert all(bs["on_periods"] == [2, 1, 2, 1])
    assert all(bs["off_periods"] == [1, 1, 5, 2])

    bs = _blink_statistics(
        locdata_with_zero_frame, memory=0, remove_heading_off_periods=True
    )
    assert all(bs["on_periods"] == [3, 1, 2, 1])
    assert all(bs["off_periods"] == [1, 5, 2])

    bs = _blink_statistics(
        locdata_without_zero_frame, memory=0, remove_heading_off_periods=True
    )
    assert all(bs["on_periods"] == [2, 1, 2, 1])
    assert all(bs["off_periods"] == [1, 5, 2])

    bs = _blink_statistics(
        locdata_with_zero_frame, memory=1, remove_heading_off_periods=False
    )
    assert all(bs["on_periods"] == [5, 2, 1])
    assert all(bs["off_periods"] == [5, 2])

    bs = _blink_statistics(
        locdata_without_zero_frame, memory=1, remove_heading_off_periods=False
    )
    assert all(bs["on_periods"] == [5, 2, 1])
    assert all(bs["off_periods"] == [5, 2])

    bs = _blink_statistics(
        locdata_with_zero_frame, memory=2, remove_heading_off_periods=False
    )
    assert all(bs["on_periods"] == [5, 5])
    assert all(bs["off_periods"] == [5])

    bs = _blink_statistics(
        locdata_without_zero_frame, memory=2, remove_heading_off_periods=False
    )
    assert all(bs["on_periods"] == [5, 5])
    assert all(bs["off_periods"] == [5])


def test_blink_statistics__with_repetitions(locdata_with_repetitions):
    _blink_statistics(
        locdata_with_repetitions, memory=0, remove_heading_off_periods=False
    )


def test_BlinkStatistics_empty(caplog):
    bs = BlinkStatistics().compute(LocData())
    bs.fit_distributions()
    bs.hist()
    assert caplog.record_tuples == [
        ("locan.analysis.blinking", 30, "Locdata is empty."),
        ("locan.analysis.blinking", 30, "No results available to fit."),
    ]


def test_BlinkStatistics(locdata_with_zero_frame):
    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    assert repr(bs) == "BlinkStatistics(memory=0, remove_heading_off_periods=True)"
    assert all(bs.results["on_periods"] == [3, 1, 2, 1])
    assert all(bs.results["off_periods"] == [1, 5, 2])
    assert bs.distribution_statistics == {}

    bs.hist(data_identifier="on_periods", ax=None, bins="auto", log=True, fit=False)
    bs.hist(data_identifier="off_periods", ax=None, bins="auto", log=True, fit=False)
    bs.hist(data_identifier="on_periods", ax=None, bins="auto", log=True, fit=True)


def test_DistributionFits(locdata_with_zero_frame):
    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    df = _DistributionFits(bs, distribution=stats.expon, data_identifier="on_periods")
    # print(df.analysis_class.results)
    assert len(df.analysis_class.results) == 5
    assert df.data_identifier == "on_periods"
    assert (
        repr(df) == "_DistributionFits(analysis_class=BlinkStatistics, "
        "distribution=expon_gen, data_identifier=on_periods)"
    )
    assert df.parameter_dict() == {}
    df.fit()
    assert list(df.parameter_dict().keys()) == ["on_periods_loc", "on_periods_scale"]

    df = _DistributionFits(bs, distribution=stats.expon, data_identifier="off_periods")
    df.fit()
    assert list(df.parameter_dict().keys()) == ["off_periods_loc", "off_periods_scale"]
    df.plot()

    # print(df.analysis_class.results[df.data_identifier])


def test_fit_distributions(locdata_with_zero_frame):
    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    bs.fit_distributions()
    assert bs.distribution_statistics["on_periods"].parameter_dict() == {
        "on_periods_loc": 1.0,
        "on_periods_scale": 0.75,
    }
    assert bs.distribution_statistics["off_periods"].parameter_dict() == {
        "off_periods_loc": 1.0,
        "off_periods_scale": 1.6666666666666665,
    }
    bs.hist()
    bs.hist(data_identifier="off_periods")
    del bs

    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    bs.fit_distributions(with_constraints=False)
    assert (
        bs.distribution_statistics["on_periods"].parameter_dict()["on_periods_loc"] == 1
    )
    assert (
        bs.distribution_statistics["off_periods"].parameter_dict()["off_periods_loc"]
        == 1
    )
    del bs

    bs = BlinkStatistics().compute(locdata_with_zero_frame)
    bs.fit_distributions(data_identifier="on_periods")
    assert bs.distribution_statistics["on_periods"].parameter_dict() == {
        "on_periods_loc": 1.0,
        "on_periods_scale": 0.75,
    }
